# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import os
import sys
import shutil
import struct
from glob import glob

import pytest
import numpy as np


@pytest.fixture
def data_dir():
    """Return path to the top-level data directory for tests."""
    parent = os.path.dirname(__file__)
    return os.path.join(parent, "data")


def make_mock_metric_db(
    parent, name, nprocs, nnodes, nthreads=1, nmetrics=2, values=[0.0, 1.0]
):
    """Create a set of mocked-up metric DB files.

    Args:
        parent (str): parent (database) directory for metric DB
        name (str): name of the experiment (e.g., lulesh2.0)
        nprocs (int): number of processes in the fake experiment
        nthreads (int): number of threads per process
        nmetrics (int): number of metrics in the metric DB
        values (list): list of float values to fill in for metric values
            on nodes

    Creates a set of ``nprocs`` x ``nthreads`` metric DB files under
    ``parent``.
    """
    if len(values) != nmetrics:
        raise ValueError("values must have length equal to nmetrics")

    # TODO: implement threads correctly.  For now, fail for threaded runs.
    assert nthreads == 1

    for p, t in np.ndindex(nprocs, nthreads):
        # TODO: generate pid and other identifiers in the filename, as well
        filename = "1.%s-%06d-%03d-a8c00270-160795-0.metric-db" % (name, p, t)
        path = os.path.join(parent, filename)

        with open(path, "wb") as f:
            f.write("HPCPROF-metricdb")  # 16 bytes
            f.write("__00.10")  # 2 bytes + 5 byte version
            f.write("b")  # 1 byte endian

            # TODO: is this really padding or is there more to
            # TODO: the format?
            f.write(8 * "\0")  # 8 bytes padding

            # write dummy values into file
            for n, m in np.ndindex(nnodes, nmetrics):
                f.write(struct.pack(">f8", values[m]))


@pytest.fixture
def calc_pi_hpct_db(data_dir, tmpdir):
    """Builds a temporary directory containing the calc-pi database."""
    hpct_db_dir = os.path.join(data_dir, "hpctoolkit-cpi-database")

    for f in glob(os.path.join(str(hpct_db_dir), "*.metric-db")):
        shutil.copy(f, str(tmpdir))
    shutil.copy(os.path.join(hpct_db_dir, "experiment.xml"), str(tmpdir))

    return tmpdir


@pytest.fixture
def osu_allgather_hpct_db(data_dir, tmpdir):
    """Builds a temporary directory containing the osu allgather database."""
    hpct_db_dir = os.path.join(data_dir, "hpctoolkit-allgather-database")

    for f in glob(os.path.join(str(hpct_db_dir), "*.metric-db")):
        shutil.copy(f, str(tmpdir))
    shutil.copy(os.path.join(hpct_db_dir, "experiment.xml"), str(tmpdir))

    return tmpdir


@pytest.fixture
def calc_pi_hpct_v4_db(data_dir, tmpdir):
    """Builds a temporary directory containing the osu allgather database."""
    hpct_db_dir = os.path.join(data_dir, "hatchet_cpi_v4_database")

    for f in glob(os.path.join(str(hpct_db_dir), "*.db")):
        shutil.copy(f, str(tmpdir))

    metrics_dir = os.path.join(str(hpct_db_dir), "metrics")
    shutil.copytree(metrics_dir, os.path.join(str(tmpdir), "metrics"))

    return tmpdir


@pytest.fixture
def hatchet_cycle_pstats(data_dir, tmpdir):
    """Builds a temporary directory containing the pstats from a profile of the hpctoolkit_reader function."""
    cprof_pstats_dir = os.path.join(data_dir, "cprofile-hatchet-pstats")
    if sys.version_info[0] == 2:
        cprof_pstats_file = os.path.join(cprof_pstats_dir, "cprofile-cycle-py2.pstats")

        shutil.copy(cprof_pstats_file, str(tmpdir))
        tmpfile = os.path.join(str(tmpdir), "cprofile-cycle-py2.pstats")
    else:
        cprof_pstats_file = os.path.join(cprof_pstats_dir, "cprofile-cycle.pstats")
        shutil.copy(cprof_pstats_file, str(tmpdir))
        tmpfile = os.path.join(str(tmpdir), "cprofile-cycle.pstats")

    return tmpfile


@pytest.fixture
def calc_pi_caliper_json(data_dir, tmpdir):
    """Builds a temporary directory containing the calc-pi JSON file."""
    cali_json_dir = os.path.join(data_dir, "caliper-cpi-json")
    cali_json_file = os.path.join(cali_json_dir, "cpi-callpath-profile.json")

    shutil.copy(cali_json_file, str(tmpdir))
    tmpfile = os.path.join(str(tmpdir), "cpi-callpath-profile.json")

    return tmpfile


@pytest.fixture
def lulesh_caliper_json(data_dir, tmpdir):
    """Builds a temporary directory containing the lulesh JSON file."""
    cali_json_dir = os.path.join(data_dir, "caliper-lulesh-json")
    cali_json_file = os.path.join(cali_json_dir, "lulesh-annotation-profile.json")

    shutil.copy(cali_json_file, str(tmpdir))
    tmpfile = os.path.join(str(tmpdir), "lulesh-annotation-profile.json")

    return tmpfile


@pytest.fixture
def lulesh_caliper_cali(data_dir, tmpdir):
    """Builds a temporary directory containing the lulesh cali file."""
    cali_dir = os.path.join(data_dir, "caliper-lulesh-cali")
    cali_file = os.path.join(cali_dir, "lulesh-annotation-profile.cali")

    shutil.copy(cali_file, str(tmpdir))
    tmpfile = os.path.join(str(tmpdir), "lulesh-annotation-profile.cali")

    return tmpfile


@pytest.fixture
def calc_pi_callgrind_dot(data_dir, tmpdir):
    """Builds a temporary directory containing the calc-pi callgrind DOT file."""
    gprof_dot_dir = os.path.join(data_dir, "gprof2dot-cpi")
    gprof_dot_file = os.path.join(gprof_dot_dir, "callgrind.dot.64042.0.1")

    shutil.copy(gprof_dot_file, str(tmpdir))
    tmpfile = os.path.join(str(tmpdir), "callgrind.dot.64042.0.1")

    return tmpfile


@pytest.fixture
def gprof_dot(data_dir, tmpdir):
    """Builds a temporary directory containing the gprof DOT file."""
    gprof_dot_dir = os.path.join(data_dir, "gprof2dot-test")
    gprof_dot_file = os.path.join(gprof_dot_dir, "gprof-profile.dot")

    shutil.copy(gprof_dot_file, str(tmpdir))
    tmpfile = os.path.join(str(tmpdir), "gprof-profile.dot")

    return tmpfile


@pytest.fixture
def hatchet_pyinstrument_json(data_dir, tmpdir):
    """Builds a temporary directory containing the pyinstrument Hatchet json file."""
    pyinstrument_json_dir = os.path.join(data_dir, "pyinstrument-hatchet-json")
    pyinstrument_json_file = os.path.join(
        pyinstrument_json_dir, "pyinstrument-hatchet-profile.json"
    )

    shutil.copy(pyinstrument_json_file, str(tmpdir))
    tmpfile = os.path.join(str(tmpdir), "pyinstrument-hatchet-profile.json")

    return tmpfile


@pytest.fixture
def tau_profile_dir(data_dir, tmpdir):
    """Builds a temporary directory containing the TAU profiles."""
    tau_dir = os.path.join(data_dir, "tau-cpi-mpi")

    for f in glob(tau_dir + "/*"):
        shutil.copy(f, str(tmpdir))

    return tau_dir


@pytest.fixture
def scorep_profile_cubex(data_dir, tmpdir):
    """Builds a temporary directory containing the Score-P profiles."""
    scorep_cubex_dir = os.path.join(data_dir, "scorep-cpi-cubex")
    scorep_cubex__file = os.path.join(scorep_cubex_dir, "profile.cubex")

    shutil.copy(scorep_cubex__file, str(tmpdir))
    tmpfile = os.path.join(str(tmpdir), "profile.cubex")

    return tmpfile


@pytest.fixture
def mock_graph_literal():
    """Creates a mock tree

    Metasyntactic variables: https://www.ietf.org/rfc/rfc3092.txt
    """
    graph_dict = [
        {
            "frame": {"name": "foo", "type": "function"},
            "metrics": {"time (inc)": 135.0, "time": 0.0},
            "children": [
                {
                    "frame": {"name": "bar"},
                    "metrics": {"time (inc)": 20.0, "time": 5.0},
                    "children": [
                        {
                            "frame": {"name": "baz", "type": "function"},
                            "metrics": {"time (inc)": 5.0, "time": 5.0},
                        },
                        {
                            "frame": {"name": "grault"},
                            "metrics": {"time (inc)": 10.0, "time": 10.0},
                        },
                    ],
                },
                {
                    "frame": {"name": "qux", "type": "function"},
                    "metrics": {"time (inc)": 60.0, "time": 0.0},
                    "children": [
                        {
                            "frame": {"name": "quux"},
                            "metrics": {"time (inc)": 60.0, "time": 5.0},
                            "children": [
                                {
                                    "frame": {"name": "corge", "type": "function"},
                                    "metrics": {"time (inc)": 55.0, "time": 10.0},
                                    "children": [
                                        {
                                            "frame": {"name": "bar"},
                                            "metrics": {
                                                "time (inc)": 20.0,
                                                "time": 5.0,
                                            },
                                            "children": [
                                                {
                                                    "frame": {
                                                        "name": "baz",
                                                        "type": "function",
                                                    },
                                                    "metrics": {
                                                        "time (inc)": 5.0,
                                                        "time": 5.0,
                                                    },
                                                },
                                                {
                                                    "frame": {"name": "grault"},
                                                    "metrics": {
                                                        "time (inc)": 10.0,
                                                        "time": 10.0,
                                                    },
                                                },
                                            ],
                                        },
                                        {
                                            "frame": {"name": "grault"},
                                            "metrics": {
                                                "time (inc)": 10.0,
                                                "time": 10.0,
                                            },
                                        },
                                        {
                                            "frame": {
                                                "name": "garply",
                                                "type": "function",
                                            },
                                            "metrics": {
                                                "time (inc)": 15.0,
                                                "time": 15.0,
                                            },
                                        },
                                    ],
                                }
                            ],
                        }
                    ],
                },
                {
                    "frame": {"name": "waldo", "type": "function"},
                    "metrics": {"time (inc)": 55.0, "time": 0.0},
                    "children": [
                        {
                            "frame": {"name": "fred", "type": "function"},
                            "metrics": {"time (inc)": 40.0, "time": 5.0},
                            "children": [
                                {
                                    "frame": {"name": "plugh", "type": "function"},
                                    "metrics": {"time (inc)": 5.0, "time": 5.0},
                                },
                                {
                                    "frame": {"name": "xyzzy", "type": "function"},
                                    "metrics": {"time (inc)": 30.0, "time": 5.0},
                                    "children": [
                                        {
                                            "frame": {
                                                "name": "thud",
                                                "type": "function",
                                            },
                                            "metrics": {
                                                "time (inc)": 25.0,
                                                "time": 5.0,
                                            },
                                            "children": [
                                                {
                                                    "frame": {
                                                        "name": "baz",
                                                        "type": "function",
                                                    },
                                                    "metrics": {
                                                        "time (inc)": 5.0,
                                                        "time": 5.0,
                                                    },
                                                },
                                                {
                                                    "frame": {
                                                        "name": "garply",
                                                        "type": "function",
                                                    },
                                                    "metrics": {
                                                        "time (inc)": 15.0,
                                                        "time": 15.0,
                                                    },
                                                },
                                            ],
                                        }
                                    ],
                                },
                            ],
                        },
                        {
                            "frame": {"name": "garply", "type": "function"},
                            "metrics": {"time (inc)": 15.0, "time": 15.0},
                        },
                    ],
                },
            ],
        },
        {
            "frame": {"name": "waldo", "type": "function"},
            "metrics": {"time (inc)": 30.0, "time": 10.0},
            "children": [
                {
                    "frame": {"name": "bar"},
                    "metrics": {"time (inc)": 20.0, "time": 5.0},
                    "children": [
                        {
                            "frame": {"name": "baz", "type": "function"},
                            "metrics": {"time (inc)": 5.0, "time": 5.0},
                        },
                        {
                            "frame": {"name": "grault"},
                            "metrics": {"time (inc)": 10.0, "time": 10.0},
                        },
                    ],
                }
            ],
        },
    ]

    return graph_dict


@pytest.fixture
def mock_dag_literal1():
    """Creates a mock DAG."""
    dag_ldict = [
        {
            "frame": {"name": "A", "type": "function"},
            "metrics": {"time (inc)": 30.0, "time": 0.0},
            "children": [
                {
                    "frame": {"name": "B", "type": "function"},
                    "metrics": {"time (inc)": 11.0, "time": 5.0},
                    "children": [
                        {
                            "frame": {"name": "C", "type": "function"},
                            "metrics": {"time (inc)": 6.0, "time": 5.0},
                            "children": [
                                {
                                    "frame": {"name": "D", "type": "function"},
                                    "metrics": {"time (inc)": 1.0, "time": 1.0},
                                }
                            ],
                        }
                    ],
                },
                {
                    "frame": {"name": "E", "type": "function"},
                    "metrics": {"time (inc)": 19.0, "time": 10.0},
                    "children": [
                        {
                            "frame": {"name": "F", "type": "function"},
                            "metrics": {"time (inc)": 9.0, "time": 9.0},
                        }
                    ],
                },
            ],
        }
    ]

    return dag_ldict


@pytest.fixture
def mock_dag_literal2():
    """Creates a mock DAG."""
    dag_ldict = [
        {
            "frame": {"name": "A", "type": "function"},
            "metrics": {"time (inc)": 30.0, "time": 0.0},
            "children": [
                {
                    "frame": {"name": "B", "type": "function"},
                    "metrics": {"time (inc)": 11.0, "time": 5.0},
                    "children": [
                        {
                            "frame": {"name": "C", "type": "function"},
                            "metrics": {"time (inc)": 6.0, "time": 5.0},
                            "children": [
                                {
                                    "frame": {"name": "D", "type": "function"},
                                    "metrics": {"time (inc)": 1.0, "time": 1.0},
                                }
                            ],
                        }
                    ],
                },
                {
                    "frame": {"name": "E", "type": "function"},
                    "metrics": {"time (inc)": 19.0, "time": 10.0},
                    "children": [
                        {
                            "frame": {"name": "H", "type": "function"},
                            "metrics": {"time (inc)": 9.0, "time": 9.0},
                        }
                    ],
                },
            ],
        }
    ]

    return dag_ldict


@pytest.fixture
def small_mock1():
    ldict = [
        {
            "frame": {"name": "A", "type": "function"},
            "metrics": {"time (inc)": 39.0, "time": 0.0},
            "children": [
                {
                    "frame": {"name": "B", "type": "function"},
                    "metrics": {"time (inc)": 10.0, "time": 5.0},
                    "children": [
                        {
                            "frame": {"name": "C", "type": "function"},
                            "metrics": {"time (inc)": 5.0, "time": 5.0},
                        }
                    ],
                },
                {
                    "frame": {"name": "E", "type": "function"},
                    "metrics": {"time (inc)": 19.0, "time": 10.0},
                    "children": [
                        {
                            "frame": {"name": "F", "type": "function"},
                            "metrics": {"time (inc)": 9.0, "time": 9.0},
                        }
                    ],
                },
                {
                    "frame": {"name": "H", "type": "function"},
                    "metrics": {"time (inc)": 10.0, "time": 10.0},
                },
            ],
        }
    ]

    return ldict


@pytest.fixture
def small_mock2():
    ldict = [
        {
            "frame": {"name": "A", "type": "function"},
            "metrics": {"time (inc)": 38.0, "time": 0.0},
            "children": [
                {
                    "frame": {"name": "B", "type": "function"},
                    "metrics": {"time (inc)": 10.0, "time": 0.0},
                    "children": [
                        {
                            "frame": {"name": "C", "type": "function"},
                            "metrics": {"time (inc)": 5.0, "time": 5.0},
                        },
                        {
                            "frame": {"name": "D", "type": "function"},
                            "metrics": {"time (inc)": 5.0, "time": 5.0},
                        },
                    ],
                },
                {
                    "frame": {"name": "E", "type": "function"},
                    "metrics": {"time (inc)": 28.0, "time": 10.0},
                    "children": [
                        {
                            "frame": {"name": "F", "type": "function"},
                            "metrics": {"time (inc)": 9.0, "time": 9.0},
                        },
                        {
                            "frame": {"name": "G", "type": "function"},
                            "metrics": {"time (inc)": 9.0, "time": 9.0},
                        },
                    ],
                },
            ],
        }
    ]

    return ldict


@pytest.fixture
def small_mock3():
    ldict = [
        {
            "frame": {"name": "A", "type": "function"},
            "metrics": {"time (inc)": 24.0, "time": 0.0},
            "children": [
                {
                    "frame": {"name": "B", "type": "function"},
                    "metrics": {"time (inc)": 5.0, "time": 5.0},
                },
                {
                    "frame": {"name": "E", "type": "function"},
                    "metrics": {"time (inc)": 19.0, "time": 10.0},
                    "children": [
                        {
                            "frame": {"name": "F", "type": "function"},
                            "metrics": {"time (inc)": 9.0, "time": 9.0},
                        }
                    ],
                },
            ],
        }
    ]

    return ldict


@pytest.fixture
def mock_dag_literal_module():
    """Creates a mock DAG."""
    dag_ldict = [
        {
            "frame": {"name": "A", "type": "function"},
            "metrics": {"time (inc)": 29.0, "time": 0.0, "module": "main"},
            "children": [
                {
                    "frame": {"name": "B", "type": "function"},
                    "metrics": {"time (inc)": 10.0, "time": 5.0, "module": "foo"},
                    "children": [
                        {
                            "frame": {"name": "C", "type": "function"},
                            "metrics": {
                                "time (inc)": 5.0,
                                "time": 5.0,
                                "module": "graz",
                            },
                        }
                    ],
                },
                {
                    "frame": {"name": "E", "type": "function"},
                    "metrics": {"time (inc)": 19.0, "time": 10.0, "module": "bar"},
                    "children": [
                        {
                            "frame": {"name": "F", "type": "function"},
                            "metrics": {
                                "time (inc)": 9.0,
                                "time": 9.0,
                                "module": "baz",
                            },
                        }
                    ],
                },
            ],
        }
    ]

    return dag_ldict


@pytest.fixture
def mock_dag_literal_module_complex():
    """Creates a mock DAG."""
    dag_ldict = [
        {
            "frame": {"name": "A", "type": "function"},
            "metrics": {"time (inc)": 5.0, "time": 1.0, "module": "main"},
            "children": [
                {
                    "frame": {"name": "B", "type": "function"},
                    "metrics": {"time (inc)": 3.0, "time": 1.0, "module": "foo"},
                    "children": [
                        {
                            "frame": {"name": "C", "type": "function"},
                            "metrics": {
                                "time (inc)": 2.0,
                                "time": 1.0,
                                "module": "graz",
                            },
                            "children": [
                                {
                                    "frame": {"name": "D", "type": "function"},
                                    "metrics": {
                                        "time (inc)": 1.0,
                                        "time": 1.0,
                                        "module": "graz",
                                    },
                                }
                            ],
                        }
                    ],
                },
                {
                    "frame": {"name": "E", "type": "function"},
                    "metrics": {"time (inc)": 1.0, "time": 1.0, "module": "bar"},
                },
            ],
        }
    ]

    return dag_ldict


@pytest.fixture
def mock_dag_literal_module_more_complex():
    """Creates a mock DAG."""
    dag_ldict = [
        {
            "frame": {"name": "A", "type": "function"},
            "metrics": {"time (inc)": 22.0, "time": 0.0, "module": "main"},
            "children": [
                {
                    "frame": {"name": "B", "type": "function"},
                    "metrics": {"time (inc)": 11.0, "time": 5.0, "module": "foo"},
                    "children": [
                        {
                            "frame": {"name": "C", "type": "function"},
                            "metrics": {
                                "time (inc)": 6.0,
                                "time": 5.0,
                                "module": "graz",
                            },
                            "children": [
                                {
                                    "frame": {"name": "D", "type": "function"},
                                    "metrics": {
                                        "time (inc)": 1.0,
                                        "time": 1.0,
                                        "module": "graz",
                                    },
                                }
                            ],
                        }
                    ],
                },
                {
                    "frame": {"name": "E", "type": "function"},
                    "metrics": {"time (inc)": 11.0, "time": 10.0, "module": "bar"},
                    "children": [
                        {
                            "frame": {"name": "F", "type": "function"},
                            "metrics": {
                                "time (inc)": 1.0,
                                "time": 1.0,
                                "module": "foo",
                            },
                        }
                    ],
                },
            ],
        }
    ]

    return dag_ldict


@pytest.fixture
def mock_graph_literal_duplicates():
    """Creates a mock tree with duplicate nodes."""
    graph_dict = [
        {
            "frame": {"name": "a", "type": "function"},
            "metrics": {"time (inc)": 130.0, "time": 0.0, "_hatchet_nid": 0},
            "children": [
                {
                    "frame": {"name": "b", "type": "function"},
                    "metrics": {"time (inc)": 20.0, "time": 5.0, "_hatchet_nid": 1},
                    "children": [
                        {
                            "frame": {"name": "d", "type": "function"},
                            "metrics": {
                                "time (inc)": 20.0,
                                "time": 5.0,
                                "_hatchet_nid": 2,
                            },
                            "children": [
                                {
                                    "frame": {"name": "e", "type": "function"},
                                    "metrics": {
                                        "time (inc)": 20.0,
                                        "time": 5.0,
                                        "_hatchet_nid": 3,
                                    },
                                },
                                {
                                    "frame": {"name": "f", "type": "function"},
                                    "metrics": {
                                        "time (inc)": 20.0,
                                        "time": 5.0,
                                        "_hatchet_nid": 4,
                                    },
                                },
                            ],
                        }
                    ],
                },
                {
                    "frame": {"name": "c", "type": "function"},
                    "metrics": {"time (inc)": 20.0, "time": 5.0, "_hatchet_nid": 5},
                    "children": [
                        {
                            "frame": {"name": "a", "type": "function"},
                            "metrics": {
                                "time (inc)": 130.0,
                                "time": 5.0,
                                "_hatchet_nid": 0,
                            },
                        },
                        {
                            "frame": {"name": "d", "type": "function"},
                            "metrics": {
                                "time (inc)": 20.0,
                                "time": 5.0,
                                "_hatchet_nid": 2,
                            },
                        },
                    ],
                },
            ],
        }
    ]

    return graph_dict


@pytest.fixture
def mock_graph_literal_duplicate_first():
    """Creates a mock tree with node with duplicate first."""
    """Creates a mock tree with duplicate nodes."""
    graph_dict = [
        {
            "frame": {"name": "a", "type": "function"},
            "metrics": {"time (inc)": 130.0, "time": 0.0, "_hatchet_nid": 0},
            "children": [
                {
                    "frame": {"name": "b", "type": "function"},
                    "metrics": {"time (inc)": 20.0, "time": 5.0, "_hatchet_nid": 1},
                    "children": [
                        {
                            "frame": {"name": "d", "type": "function"},
                            "metrics": {
                                "time (inc)": 20.0,
                                "time": 5.0,
                                "_hatchet_nid": 2,
                            },
                            "children": [
                                {
                                    "frame": {"name": "e", "type": "function"},
                                    "metrics": {
                                        "time (inc)": 20.0,
                                        "time": 5.0,
                                        "_hatchet_nid": 3,
                                    },
                                },
                                {
                                    "frame": {"name": "f", "type": "function"},
                                    "metrics": {
                                        "time (inc)": 20.0,
                                        "time": 5.0,
                                        "_hatchet_nid": 4,
                                    },
                                },
                            ],
                        }
                    ],
                },
                {
                    "frame": {"name": "c", "type": "function"},
                    "metrics": {"time (inc)": 20.0, "time": 5.0, "_hatchet_nid": 5},
                    "children": [
                        {
                            "frame": {"name": "a", "type": "function"},
                            "metrics": {
                                "time (inc)": 130.0,
                                "time": 5.0,
                                "_hatchet_nid": 0,
                            },
                        },
                        {
                            "frame": {"name": "d", "type": "function"},
                            "metrics": {
                                "time (inc)": 20.0,
                                "time": 5.0,
                                "_hatchet_nid": 2,
                            },
                        },
                    ],
                },
            ],
        }
    ]

    return graph_dict


@pytest.fixture
def timemory_json_data():
    import numpy as np
    import timemory
    from timemory.bundle import marker
    from timemory.trace import Config as TracerConfig
    from timemory.profiler import Config as ProfilerConfig
    from timemory_func import prof_func, trace_func, components

    # disable automatic output during finalization
    timemory.settings.auto_output = False

    # enable flat collection because of the coverage exe
    timemory.settings.flat_profile = True

    with marker(components, key="main"):
        nx = 10
        ny = 10
        tol = 5.0e-2
        profl_arr = np.random.rand(nx, ny)
        trace_arr = np.zeros([nx, ny], dtype=np.float64)
        trace_arr[:, :] = profl_arr[:, :]

        # restrict the scope of the profiler
        ProfilerConfig.only_filenames = ["timemory_func.py", "_methods.py"]
        ProfilerConfig.include_line = False
        prof_func(profl_arr, tol)

        # restrict the scope of the tracer
        TracerConfig.only_filenames = ["timemory_func.py"]
        trace_func(trace_arr, tol)

    return timemory.get(hierarchy=True)


@pytest.fixture
def mock_graph_inc_metric_only():
    ldict = [
        {
            "frame": {"name": "A", "type": "function"},
            "metrics": {"time (inc)": 130.0},
            "children": [
                {
                    "frame": {"name": "B", "type": "function"},
                    "metrics": {"time (inc)": 20.0},
                    "children": [
                        {
                            "frame": {"name": "C", "type": "function"},
                            "metrics": {"time (inc)": 5.0},
                        }
                    ],
                },
                {
                    "frame": {"name": "E", "type": "function"},
                    "metrics": {"time (inc)": 55.0},
                    "children": [
                        {
                            "frame": {"name": "F", "type": "function"},
                            "metrics": {"time (inc)": 1.0},
                        }
                    ],
                },
                {
                    "frame": {"name": "H", "type": "function"},
                    "metrics": {"time (inc)": 55.0},
                },
            ],
        }
    ]

    return ldict


@pytest.fixture
def spotdb_data(data_dir):
    import spotdb

    db = spotdb.connect(os.path.join(data_dir, "spotdb"))
    return db


@pytest.fixture
def sw4_caliper_cuda_activity_profile_cali(data_dir, tmpdir):
    cali_json_dir = os.path.join(data_dir, "caliper-sw4-cuda-activity-cali")
    cali_json_file = os.path.join(cali_json_dir, "caliper_cuda_activity_profile.cali")

    shutil.copy(cali_json_file, str(tmpdir))
    tmpfile = os.path.join(str(tmpdir), "caliper_cuda_activity_profile.cali")

    return tmpfile


@pytest.fixture
def sw4_caliper_cuda_activity_profile_summary_cali(data_dir, tmpdir):
    cali_json_dir = os.path.join(data_dir, "caliper-sw4-cuda-activity-profile-cali")
    cali_json_file = os.path.join(
        cali_json_dir, "caliper_cuda_activity_profile_summary_v2.cali"
    )

    shutil.copy(cali_json_file, str(tmpdir))
    tmpfile = os.path.join(str(tmpdir), "caliper_cuda_activity_profile_summary_v2.cali")

    return tmpfile


@pytest.fixture
def omnitrace_python_source(data_dir, tmpdir):
    """single rank, single thread"""
    omnitrace_data_dir = os.path.join(data_dir, "perfetto")
    omnitrace_data_file = "python-source.proto"

    shutil.copy(os.path.join(omnitrace_data_dir, omnitrace_data_file), str(tmpdir))
    tmpfile = os.path.join(str(tmpdir), omnitrace_data_file)

    return tmpfile


@pytest.fixture
def omnitrace_mpi_aggregate(data_dir, tmpdir):
    """4 ranks each with 2 threads in one file"""
    omnitrace_data_dir = os.path.join(data_dir, "perfetto")
    omnitrace_data_file = "mpi-binary-rewrite-all.proto"

    shutil.copy(os.path.join(omnitrace_data_dir, omnitrace_data_file), str(tmpdir))
    tmpfile = os.path.join(str(tmpdir), omnitrace_data_file)

    return tmpfile


@pytest.fixture
def omnitrace_mpi_single(data_dir, tmpdir):
    """1 rank with 2 threads in one file"""
    omnitrace_data_dir = os.path.join(data_dir, "perfetto")
    omnitrace_data_file = "mpi-binary-rewrite-0.proto"

    shutil.copy(os.path.join(omnitrace_data_dir, omnitrace_data_file), str(tmpdir))
    tmpfile = os.path.join(str(tmpdir), omnitrace_data_file)

    return tmpfile


@pytest.fixture
def omnitrace_mpi_group(data_dir, tmpdir):
    """4 ranks each with 2 threads in 4 files"""
    omnitrace_data_dir = os.path.join(data_dir, "perfetto")
    omnitrace_data_file = "mpi-binary-rewrite-{}.proto"

    for n in range(4):
        shutil.copy(
            os.path.join(omnitrace_data_dir, omnitrace_data_file.format(n)), str(tmpdir)
        )

    tmpfile = os.path.join(str(tmpdir), omnitrace_data_file.format("*"))

    return tmpfile



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from abc import abstractmethod

try:
    from abc import ABC
except ImportError:
    from abc import ABCMeta

    ABC = ABCMeta("ABC", (object,), {"__slots__": ()})

from itertools import groupby
from numbers import Real
import re
import sys
import pandas as pd
from pandas import DataFrame
from pandas.core.indexes.multi import MultiIndex
from textx import metamodel_from_str
from textx.exceptions import TextXError

# Flake8 to ignore this import, it does not recognize that eval("np.nan") needs
# numpy package
import numpy as np  # noqa: F401

from .node import Node, traversal_order


class AbstractQuery(ABC):
    """Abstract Base Class defining a Hatchet Query"""

    @abstractmethod
    def apply(self, gf):
        """Apply the query to a GraphFrame.

        Arguments:
            gf (GraphFrame): the GraphFrame on which to apply the query.

        Returns:
            (list): A list representing the set of nodes from paths that match this query.
        """
        pass

    def __and__(self, other):
        """Create an AndQuery with this query and another.

        Arguments:
            other (GraphFrame): the other query to use in the AndQuery.

        Returns:
            (AndQuery): A query object representing the intersection of the two queries.
        """
        return AndQuery(self, other)

    def __or__(self, other):
        """Create an OrQuery with this query and another.

        Arguments:
            other (GraphFrame): the other query to use in the OrQuery.

        Returns:
            (OrQuery): A query object representing the union of the two queries.
        """
        return OrQuery(self, other)

    def __xor__(self, other):
        """Create a XorQuery with this query and another.

        Arguments:
            other (GraphFrame): the other query to use in the XorQuery.

        Returns:
            (XorQuery): A query object representing the symmetric difference of the two queries.
        """
        return XorQuery(self, other)

    def __invert__(self):
        """Create a NotQuery with this query.

        Returns:
            (NotQuery): A query object representing all nodes that don't match this query.
        """
        return NotQuery(self)


class NaryQuery(AbstractQuery):
    """Abstract Base Class defining a compound query
    that acts on and merges N separate subqueries"""

    def __init__(self, *args):
        """Create a new NaryQuery object.

        Arguments:
            *args (tuple): the subqueries (high-level, low-level, or compound) to be performed.
        """
        self.subqueries = []
        if isinstance(args[0], tuple) and len(args) == 1:
            args = args[0]
        for query in args:
            if isinstance(query, list):
                self.subqueries.append(QueryMatcher(query))
            elif isinstance(query, str):
                self.subqueries.append(CypherQuery(query))
            elif issubclass(type(query), AbstractQuery):
                self.subqueries.append(query)
            else:
                raise TypeError(
                    "Subqueries for NaryQuery must be either a \
                                high-level query or a subclass of AbstractQuery"
                )

    @abstractmethod
    def _perform_nary_op(self, query_results, gf):
        """Perform the NaryQuery subclass's designated operation on the results of the subqueries.

        Arguments:
            query_results (list): the results of the subqueries.
            gf (GraphFrame): the GraphFrame on which the query is applied.

        Returns:
            (list): A list of nodes representing the result of applying the subclass-designated operation to the results of the subqueries.
        """
        pass

    def apply(self, gf):
        """Apply the NaryQuery to a GraphFrame.

        Arguments:
            gf (GraphFrame): the GraphFrame on which to apply the query.

        Returns:
            (list): A list of nodes representing the result of applying the subclass-designated operation to the results of the subqueries.
        """
        results = []
        for query in self.subqueries:
            results.append(query.apply(gf))
        return self._perform_nary_op(results, gf)


class QueryMatcher(AbstractQuery):
    """Process and apply queries to GraphFrames."""

    def __init__(self, query=None):
        """Create a new QueryMatcher object.

        Arguments:
            query (list, optional): if provided, convert the contents of the high-level API query into an internal representation.
        """
        # Initialize containers for query and memoization cache.
        self.query_pattern = []
        self.search_cache = {}
        # If a high-level API list is provided, process it.
        if query is not None:
            assert isinstance(query, list)

            def _convert_dict_to_filter(attr_filter):
                """Converts high-level API attribute filter to a lambda"""
                compops = ("<", ">", "==", ">=", "<=", "<>", "!=")  # ,
                # Currently not supported
                #           "is", "is not", "in", "not in")

                # This is a dict to work around Python's non-local variable
                # assignment rules.
                #
                # TODO: Replace this with the use of the "nonlocal" keyword
                #       once Python 2.7 support is dropped.
                first_no_drop_indices = {"val": True}

                def filter_series(df_row):
                    def filter_single_series(df_row, key, single_value):
                        if key == "depth":
                            node = df_row.name
                            if isinstance(
                                single_value, str
                            ) and single_value.lower().startswith(compops):
                                return eval("{} {}".format(node._depth, single_value))
                            if isinstance(single_value, Real):
                                return node._depth == single_value
                            raise InvalidQueryFilter(
                                "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                                    key
                                )
                            )
                        if key == "node_id":
                            node = df_row.name
                            if isinstance(
                                single_value, str
                            ) and single_value.lower().startswith(compops):
                                return eval(
                                    "{} {}".format(node._hatchet_nid, single_value)
                                )
                            if isinstance(single_value, Real):
                                return node._hatchet_nid == single_value
                            raise InvalidQueryFilter(
                                "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                                    key
                                )
                            )
                        if key not in df_row.keys():
                            return False
                        if isinstance(df_row[key], str):
                            if not isinstance(single_value, str):
                                raise InvalidQueryFilter(
                                    "Value for attribute {} must be a string.".format(
                                        key
                                    )
                                )
                            return (
                                re.match(single_value + r"\Z", df_row[key]) is not None
                            )
                        if isinstance(df_row[key], Real):
                            if isinstance(
                                single_value, str
                            ) and single_value.lower().startswith(compops):
                                # compare nan metric value to numeric query
                                # (e.g. np.nan > 5)
                                if pd.isnull(df_row[key]):
                                    nan_str = "np.nan"
                                    # compare nan metric value to nan query
                                    # (e.g., np.nan == np.nan)
                                    if nan_str in single_value:
                                        return eval(
                                            "pd.isnull({}) == True".format(nan_str)
                                        )
                                    return eval("{} {}".format(nan_str, single_value))
                                elif np.isinf(df_row[key]):
                                    inf_str = "np.inf"
                                    # compare inf metric value to inf query
                                    # (e.g., np.inf == np.inf)
                                    if inf_str in single_value:
                                        return eval(
                                            "np.isinf({}) == True".format(inf_str)
                                        )
                                    return eval("{} {}".format(inf_str, single_value))
                                else:
                                    return eval(
                                        "{} {}".format(df_row[key], single_value)
                                    )

                            if isinstance(single_value, Real):
                                return df_row[key] == single_value
                            raise InvalidQueryFilter(
                                "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                                    key
                                )
                            )
                        raise InvalidQueryFilter(
                            "Filter must be one of the following:\n  * A regex string for a String attribute\n  * A string starting with a comparison operator for a Numeric attribute\n  * A number for a Numeric attribute\n"
                        )

                    matches = True
                    for k, v in attr_filter.items():
                        try:
                            _ = iter(v)
                            # Manually raise TypeError if v is a string so that
                            # the string is processed as a non-iterable
                            if isinstance(v, str):
                                raise TypeError
                        # Runs if v is not iterable (e.g., list, tuple, etc.)
                        except TypeError:
                            matches = matches and filter_single_series(df_row, k, v)
                        else:
                            for single_value in v:
                                matches = matches and filter_single_series(
                                    df_row, k, single_value
                                )
                    return matches

                def filter_dframe(df_row):
                    if first_no_drop_indices["val"]:
                        print(
                            "==================================================================="
                        )
                        print(
                            "WARNING: You are performing a query without dropping index levels."
                        )
                        print(
                            "         This is a valid operation, but it will significantly"
                        )
                        print(
                            "         increase the time it takes for this operation to complete."
                        )
                        print(
                            "         If you don't want the operation to take so long, call"
                        )
                        print("         GraphFrame.drop_index_levels() before calling")
                        print("         GraphFrame.filter()")
                        print(
                            "===================================================================\n"
                        )
                        first_no_drop_indices["val"] = False

                    def filter_single_dframe(node, df_row, key, single_value):
                        if key == "depth":
                            if isinstance(
                                single_value, str
                            ) and single_value.lower().startswith(compops):
                                return eval("{} {}".format(node._depth, single_value))
                            if isinstance(single_value, Real):
                                return node._depth == single_value
                            raise InvalidQueryFilter(
                                "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                                    key
                                )
                            )
                        if key == "node_id":
                            if isinstance(
                                single_value, str
                            ) and single_value.lower().startswith(compops):
                                return eval(
                                    "{} {}".format(node._hatchet_nid, single_value)
                                )
                            if isinstance(single_value, Real):
                                return node._hatchet_nid == single_value
                            raise InvalidQueryFilter(
                                "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                                    key
                                )
                            )
                        if key not in df_row.columns:
                            return False
                        if df_row[key].apply(type).eq(str).all():
                            if not isinstance(single_value, str):
                                raise InvalidQueryFilter(
                                    "Value for attribute {} must be a string.".format(
                                        key
                                    )
                                )
                            return (
                                df_row[key]
                                .apply(
                                    lambda x: re.match(single_value + r"\Z", x)
                                    is not None
                                )
                                .any()
                            )
                        if df_row[key].apply(type).eq(Real).all():
                            if isinstance(
                                single_value, str
                            ) and single_value.lower().startswith(compops):
                                return (
                                    df_row[key]
                                    .apply(
                                        lambda x: eval("{} {}".format(x, single_value))
                                    )
                                    .any()
                                )
                            if isinstance(single_value, Real):
                                return (
                                    df_row[key].apply(lambda x: x == single_value).any()
                                )
                            raise InvalidQueryFilter(
                                "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                                    key
                                )
                            )
                        raise InvalidQueryFilter(
                            "Filter must be one of the following:\n  * A regex string for a String attribute\n  * A string starting with a comparison operator for a Numeric attribute\n  * A number for a Numeric attribute\n"
                        )

                    matches = True
                    node = df_row.name.to_frame().index[0][0]
                    for k, v in attr_filter.items():
                        try:
                            _ = iter(v)
                            if isinstance(v, str):
                                raise TypeError
                        except TypeError:
                            matches = matches and filter_single_dframe(
                                node, df_row, k, v
                            )
                        else:
                            for single_value in v:
                                matches = matches and filter_single_dframe(
                                    node, df_row, k, single_value
                                )
                    return matches

                def filter_choice(df_row):
                    if isinstance(df_row, DataFrame):
                        return filter_dframe(df_row)
                    return filter_series(df_row)

                return filter_choice if attr_filter != {} else lambda row: True

            for elem in query:
                if isinstance(elem, dict):
                    self._add_node(".", _convert_dict_to_filter(elem))
                elif isinstance(elem, str) or isinstance(elem, int):
                    self._add_node(elem)
                elif isinstance(elem, tuple):
                    assert isinstance(elem[1], dict)
                    if isinstance(elem[0], str) or isinstance(elem[0], int):
                        self._add_node(elem[0], _convert_dict_to_filter(elem[1]))
                    else:
                        raise InvalidQueryPath(
                            "The first value of a tuple entry in a path must be either a string or integer."
                        )
                else:
                    raise InvalidQueryPath(
                        "A query path must be a list containing String, Integer, Dict, or Tuple elements"
                    )

    def match(self, wildcard_spec=".", filter_func=lambda row: True):
        """Start a query with a root node described by the arguments.

        Arguments:
            wildcard_spec (str, optional, ".", "*", or "+"): the wildcard status of the node (follows standard Regex syntax)
            filter_func (callable, optional): a callable accepting only a row from a Pandas DataFrame that is used to filter this node in the query

        Returns:
            (QueryMatcher): The instance of the class that called this function (enables fluent design).
        """
        if len(self.query_pattern) != 0:
            self.query_pattern = []
        self._add_node(wildcard_spec, filter_func)
        return self

    def rel(self, wildcard_spec=".", filter_func=lambda row: True):
        """Add another edge and node to the query.

        Arguments:
            wildcard_spec (str, optional, ".", "*", or "+"): the wildcard status of the node (follows standard Regex syntax)
            filter_func (callable, optional): a callable accepting only a row from a Pandas DataFrame that is used to filter this node in the query

        Returns:
            (QueryMatcher): The instance of the class that called this function (enables fluent design).
        """
        self._add_node(wildcard_spec, filter_func)
        return self

    def apply(self, gf):
        """Apply the query to a GraphFrame.

        Arguments:
            gf (GraphFrame): the GraphFrame on which to apply the query.

        Returns:
            (list): A list representing the set of nodes from paths that match this query.
        """
        self.search_cache = {}
        matches = []
        visited = set()
        for root in sorted(gf.graph.roots, key=traversal_order):
            self._apply_impl(gf, root, visited, matches)
        assert len(visited) == len(gf.graph)
        matched_node_set = list(set().union(*matches))
        # return matches
        return matched_node_set

    def _add_node(self, wildcard_spec=".", filter_func=lambda row: True):
        """Add a node to the query.
        Arguments:
            wildcard_spec (str, optional, ".", "*", or "+"): the wildcard status of the node (follows standard Regex syntax)
            filter_func (callable, optional): a callable accepting only a row from a Pandas DataFrame that is used to filter this node in the query
        """
        assert isinstance(wildcard_spec, int) or isinstance(wildcard_spec, str)
        assert callable(filter_func)
        if isinstance(wildcard_spec, int):
            for i in range(wildcard_spec):
                self.query_pattern.append((".", filter_func))
        else:
            assert wildcard_spec == "." or wildcard_spec == "*" or wildcard_spec == "+"
            self.query_pattern.append((wildcard_spec, filter_func))

    def _cache_node(self, gf, node):
        """Cache (Memoize) the parts of the query that the node matches.

        Arguments:
            gf (GraphFrame): the GraphFrame containing the node to be cached.
            node (Node): the Node to be cached.
        """
        assert isinstance(node, Node)
        matches = []
        # Applies each filtering function to the node to cache which
        # query nodes the current node matches.
        for i, node_query in enumerate(self.query_pattern):
            _, filter_func = node_query
            row = None
            if isinstance(gf.dataframe.index, MultiIndex):
                row = pd.concat([gf.dataframe.loc[node]], keys=[node], names=["node"])
            else:
                row = gf.dataframe.loc[node]
            if filter_func(row):
                matches.append(i)
        self.search_cache[node._hatchet_nid] = matches

    def _match_0_or_more(self, gf, node, wcard_idx):
        """Process a "*" wildcard in the query on a subgraph.

        Arguments:
            gf (GraphFrame): the GraphFrame being queried.
            node (Node): the node being queried against the "*" wildcard.
            wcard_idx (int): the index associated with the "*" wildcard query.

        Returns:
            (list): a list of lists representing the paths rooted at "node" that match the "*" wildcard and/or the next query node. Will return None if there is no match for the "*" wildcard or the next query node.
        """
        # Cache the node if it's not already cached
        if node._hatchet_nid not in self.search_cache:
            self._cache_node(gf, node)
        # If the node matches with the next non-wildcard query node,
        # end the recursion and return the node.
        if wcard_idx + 1 in self.search_cache[node._hatchet_nid]:
            return [[]]
        # If the node matches the "*" wildcard query, recursively
        # apply this function to the current node's children. Then,
        # collect their returned matches, and prepend the current node.
        elif wcard_idx in self.search_cache[node._hatchet_nid]:
            matches = []
            if len(node.children) == 0:
                if wcard_idx == len(self.query_pattern) - 1:
                    return [[node]]
                return None
            for child in sorted(node.children, key=traversal_order):
                sub_match = self._match_0_or_more(gf, child, wcard_idx)
                if sub_match is not None:
                    matches.extend(sub_match)
            if len(matches) == 0:
                return None
            tmp = set(tuple(m) for m in matches)
            matches = [list(t) for t in tmp]
            return [[node] + m for m in matches]
        # If the current node doesn't match the current "*" wildcard or
        # the next non-wildcard query node, return None.
        else:
            if wcard_idx == len(self.query_pattern) - 1:
                return [[]]
            return None

    def _match_1_or_more(self, gf, node, wcard_idx):
        """Process a "+" wildcard in the query on a subgraph.

        Arguments:
            gf (GraphFrame): the GraphFrame being queried.
            node (Node): the node being queried against the "+" wildcard.
            wcard_idx (int): the index associated with the "+" wildcard query.

        Returns:
            (list): a list of lists representing the paths rooted at "node" that match the "+" wildcard and/or the next query node. Will return None if there is no match for the "+" wildcard or the next query node.
        """
        # Cache the node if it's not already cached
        if node._hatchet_nid not in self.search_cache:
            self._cache_node(gf, node)
        # If the current node doesn't match the "+" wildcard, return None.
        if wcard_idx not in self.search_cache[node._hatchet_nid]:
            return None
        # Since a query can't end on a wildcard, return None if the
        # current node has no children.
        if len(node.children) == 0:
            return None
        # Use _match_0_or_more to collect all additional wildcard matches.
        matches = []
        for child in sorted(node.children, key=traversal_order):
            sub_match = self._match_0_or_more(gf, child, wcard_idx)
            if sub_match is not None:
                matches.extend(sub_match)
        # Since _match_0_or_more will capture the query node that follows
        # the wildcard, if no paths were retrieved from that function,
        # the pattern does not continue after the "+" wildcard. Thus,
        # since a pattern cannot end on a wildcard, the function
        # returns None.
        if len(matches) == 0:
            return None
        return [[node] + m for m in matches]

    def _match_1(self, gf, node, idx):
        """Process a "." wildcard in the query on a subgraph.

        Arguments:
            gf (GraphFrame): the GraphFrame being queried.
            node (Node): the node being queried against the "." wildcard.
            wcard_idx (int): the index associated with the "." wildcard query.

        Returns:
            (list): A list of lists representing the children of "node" that match the "." wildcard being considered. Will return None if there are no matches for the "." wildcard.
        """
        if node._hatchet_nid not in self.search_cache:
            self._cache_node(gf, node)
        matches = []
        for child in sorted(node.children, key=traversal_order):
            # Cache the node if it's not already cached
            if child._hatchet_nid not in self.search_cache:
                self._cache_node(gf, child)
            if idx in self.search_cache[child._hatchet_nid]:
                matches.append([child])
        # To be consistent with the other matching functions, return
        # None instead of an empty list.
        if len(matches) == 0:
            return None
        return matches

    def _match_pattern(self, gf, pattern_root, match_idx):
        """Try to match the query pattern starting at the provided root node.

        Arguments:
            gf (GraphFrame): the GraphFrame being queried.
            pattern_root (Node): the root node of the subgraph that is being queried.

        Returns:
            (list): A list of lists representing the paths rooted at "pattern_root" that match the query.
        """
        assert isinstance(pattern_root, Node)
        # Starting query node
        pattern_idx = match_idx + 1
        if (
            self.query_pattern[match_idx][0] == "*"
            or self.query_pattern[match_idx][0] == "+"
        ):
            pattern_idx = 0
        # Starting matching pattern
        matches = [[pattern_root]]
        while pattern_idx < len(self.query_pattern):
            # Get the wildcard type
            wcard, _ = self.query_pattern[pattern_idx]
            new_matches = []
            # Consider each existing match individually so that more
            # nodes can be added to them.
            for m in matches:
                sub_match = []
                # Get the portion of the subgraph that matches the next
                # part of the query.
                if wcard == ".":
                    s = self._match_1(gf, m[-1], pattern_idx)
                    if s is None:
                        sub_match.append(s)
                    else:
                        sub_match.extend(s)
                elif wcard == "*":
                    if len(m[-1].children) == 0:
                        sub_match.append([])
                    else:
                        for child in sorted(m[-1].children, key=traversal_order):
                            s = self._match_0_or_more(gf, child, pattern_idx)
                            if s is None:
                                sub_match.append(s)
                            else:
                                sub_match.extend(s)
                elif wcard == "+":
                    if len(m[-1].children) == 0:
                        sub_match.append(None)
                    else:
                        for child in sorted(m[-1].children, key=traversal_order):
                            s = self._match_1_or_more(gf, child, pattern_idx)
                            if s is None:
                                sub_match.append(s)
                            else:
                                sub_match.extend(s)
                else:
                    raise InvalidQueryFilter(
                        'Query wildcards must be one of ".", "*", or "+"'
                    )
                # Merge the next part of the match path with the
                # existing part.
                for s in sub_match:
                    if s is not None:
                        new_matches.append(m + s)
                new_matches = [uniq_match for uniq_match, _ in groupby(new_matches)]
            # Overwrite the old matches with the updated matches
            matches = new_matches
            # If all the existing partial matches were not able to be
            # expanded into full matches, return None.
            if len(matches) == 0:
                return None
            # Update the query node
            pattern_idx += 1
        return matches

    def _apply_impl(self, gf, node, visited, matches):
        """Traverse the subgraph with the specified root, and collect all paths that match the query.

        Arguments:
            gf (GraphFrame): the GraphFrame being queried.
            node (Node): the root node of the subgraph that is being queried.
            visited (set): a set that keeps track of what nodes have been visited in the traversal to minimize the amount of work that is repeated.
            matches (list): the list in which the final set of matches are stored.
        """
        # If the node has already been visited (or is None for some
        # reason), skip it.
        if node is None or node._hatchet_nid in visited:
            return
        # Cache the node if it's not already cached
        if node._hatchet_nid not in self.search_cache:
            self._cache_node(gf, node)
        # If the node matches the starting/root node of the query,
        # try to get all query matches in the subgraph rooted at
        # this node.
        if self.query_pattern[0][0] == "*":
            if 1 in self.search_cache[node._hatchet_nid]:
                sub_match = self._match_pattern(gf, node, 1)
                if sub_match is not None:
                    matches.extend(sub_match)
        if 0 in self.search_cache[node._hatchet_nid]:
            sub_match = self._match_pattern(gf, node, 0)
            if sub_match is not None:
                matches.extend(sub_match)
        # Note that the node is now visited.
        visited.add(node._hatchet_nid)
        # Continue the Depth First Search.
        for child in sorted(node.children, key=traversal_order):
            self._apply_impl(gf, child, visited, matches)


GRAMMAR = """
FullQuery: path_expr=MatchExpr(cond_expr=WhereExpr)?;
MatchExpr: 'MATCH' path=PathQuery;
PathQuery: '(' nodes=NodeExpr ')'('->' '(' nodes=NodeExpr ')')*;
NodeExpr: ((wcard=INT | wcard=STRING) ',' name=ID) | (wcard=INT | wcard=STRING) |  name=ID;
WhereExpr: 'WHERE' ConditionExpr;
ConditionExpr: conditions+=CompoundCond;
CompoundCond: UnaryCond | BinaryCond;
BinaryCond: AndCond | OrCond;
AndCond: 'AND' subcond=UnaryCond;
OrCond: 'OR' subcond=UnaryCond;
UnaryCond: NotCond | SingleCond;
NotCond: 'NOT' subcond=SingleCond;
SingleCond: StringCond | NumberCond | NoneCond | NotNoneCond;
NoneCond: name=ID '.' prop=STRING 'IS NONE';
NotNoneCond: name=ID '.' prop=STRING 'IS NOT NONE';
StringCond: StringEq | StringStartsWith | StringEndsWith | StringContains | StringMatch;
StringEq: name=ID '.' prop=STRING '=' val=STRING;
StringStartsWith: name=ID '.' prop=STRING 'STARTS WITH' val=STRING;
StringEndsWith: name=ID '.' prop=STRING 'ENDS WITH' val=STRING;
StringContains: name=ID '.' prop=STRING 'CONTAINS' val=STRING;
StringMatch: name=ID '.' prop=STRING '=~' val=STRING;
NumberCond: NumEq | NumLt | NumGt | NumLte | NumGte | NumNan | NumNotNan | NumInf | NumNotInf;
NumEq: name=ID '.' prop=STRING '=' val=NUMBER;
NumLt: name=ID '.' prop=STRING '<' val=NUMBER;
NumGt: name=ID '.' prop=STRING '>' val=NUMBER;
NumLte: name=ID '.' prop=STRING '<=' val=NUMBER;
NumGte: name=ID '.' prop=STRING '>=' val=NUMBER;
NumNan: name=ID '.' prop=STRING 'IS NAN';
NumNotNan: name=ID '.' prop=STRING 'IS NOT NAN';
NumInf: name=ID '.' prop=STRING 'IS INF';
NumNotInf: name=ID '.' prop=STRING 'IS NOT INF';
"""

mm = metamodel_from_str(GRAMMAR)


def cname(obj):
    return obj.__class__.__name__


def filter_check_types(type_check, df_row, filt_lambda):
    try:
        if type_check == "" or eval(type_check):
            return filt_lambda(df_row)
        else:
            raise InvalidQueryFilter("Type mismatch in filter")
    except KeyError:
        return False


class CypherQuery(QueryMatcher):
    def __init__(self, cypher_query):
        if sys.version_info[0] == 2:
            super(CypherQuery, self).__init__()
        else:
            super().__init__()
        model = None
        try:
            model = mm.model_from_str(cypher_query)
        except TextXError as e:
            # TODO Change to a "raise-from" expression when Python 2.7 support is dropped
            raise InvalidQueryPath(
                'Invalid "Cypher" Query Detected. Parser Error Message: {}'.format(
                    e.message
                )
            )
        self.wcards = []
        self.wcard_pos = {}
        self._parse_path(model.path_expr)
        self.filters = [[] for _ in self.wcards]
        self._parse_conditions(model.cond_expr)
        self.lambda_filters = [None for _ in self.wcards]
        self._build_lambdas()
        self._build_query()

    def _build_query(self):
        for i in range(0, len(self.wcards)):
            wcard = self.wcards[i][0]
            # TODO Remove this when Python 2.7 support is dropped.
            if sys.version_info[0] == 2 and not isinstance(wcard, Real):
                wcard = wcard.encode("ascii", "ignore")
            filt_str = self.lambda_filters[i]
            if filt_str is None:
                if i == 0:
                    self.match(wildcard_spec=wcard)
                else:
                    self.rel(wildcard_spec=wcard)
            else:
                if i == 0:
                    self.match(wildcard_spec=wcard, filter_func=eval(filt_str))
                else:
                    self.rel(wildcard_spec=wcard, filter_func=eval(filt_str))

    def _build_lambdas(self):
        for i in range(0, len(self.wcards)):
            n = self.wcards[i]
            if n[1] != "":
                bool_expr = ""
                type_check = ""
                for j, cond in enumerate(self.filters[i]):
                    if cond[0] is not None:
                        bool_expr += " {}".format(cond[0])
                    bool_expr += " {}".format(cond[1])
                    if cond[2] is not None:
                        if j == 0:
                            type_check += " {}".format(cond[2])
                        else:
                            type_check += " and {}".format(cond[2])
                bool_expr = "lambda df_row: {}".format(bool_expr)
                bool_expr = (
                    'lambda df_row: filter_check_types("{}", df_row, {})'.format(
                        type_check, bool_expr
                    )
                )
                self.lambda_filters[i] = bool_expr

    def _parse_path(self, path_obj):
        nodes = path_obj.path.nodes
        idx = len(self.wcards)
        for n in nodes:
            new_node = [n.wcard, n.name]
            if n.wcard is None or n.wcard == "" or n.wcard == 0:
                new_node[0] = "."
            self.wcards.append(new_node)
            if n.name != "":
                self.wcard_pos[n.name] = idx
            idx += 1

    def _parse_conditions(self, cond_expr):
        conditions = cond_expr.conditions
        for cond in conditions:
            converted_condition = None
            if self._is_unary_cond(cond):
                converted_condition = self._parse_unary_cond(cond)
            elif self._is_binary_cond(cond):
                converted_condition = self._parse_binary_cond(cond)
            else:
                raise RuntimeError("Bad Condition")
            self.filters[self.wcard_pos[converted_condition[1]]].append(
                [converted_condition[0], converted_condition[2], converted_condition[3]]
            )
        for i in range(0, len(self.filters)):
            if len(self.filters[i]) > 0:
                if self.filters[i][0][0] != "not":
                    self.filters[i][0][0] = None

    def _is_unary_cond(self, obj):
        if (
            cname(obj) == "NotCond"
            or self._is_str_cond(obj)
            or self._is_num_cond(obj)
            or cname(obj) in ["NoneCond", "NotNoneCond"]
        ):
            return True
        return False

    def _is_binary_cond(self, obj):
        if cname(obj) in ["AndCond", "OrCond"]:
            return True
        return False

    def _parse_binary_cond(self, obj):
        if cname(obj) == "AndCond":
            return self._parse_and_cond(obj)
        if cname(obj) == "OrCond":
            return self._parse_or_cond(obj)
        raise RuntimeError("Bad Binary Condition")

    def _parse_or_cond(self, obj):
        converted_subcond = self._parse_unary_cond(obj.subcond)
        converted_subcond[0] = "or"
        return converted_subcond

    def _parse_and_cond(self, obj):
        converted_subcond = self._parse_unary_cond(obj.subcond)
        converted_subcond[0] = "and"
        return converted_subcond

    def _parse_unary_cond(self, obj):
        if cname(obj) == "NotCond":
            return self._parse_not_cond(obj)
        return self._parse_single_cond(obj)

    def _parse_not_cond(self, obj):
        converted_subcond = self._parse_single_cond(obj.subcond)
        converted_subcond[2] = "not {}".format(converted_subcond[2])
        return converted_subcond

    def _parse_single_cond(self, obj):
        if self._is_str_cond(obj):
            return self._parse_str(obj)
        if self._is_num_cond(obj):
            return self._parse_num(obj)
        if cname(obj) == "NoneCond":
            return self._parse_none(obj)
        if cname(obj) == "NotNoneCond":
            return self._parse_not_none(obj)
        raise RuntimeError("Bad Single Condition")

    def _parse_none(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "df_row.name._depth is None",
                None,
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid is None",
                None,
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] is None'.format(obj.prop),
            None,
        ]

    def _parse_not_none(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "df_row.name._depth is not None",
                None,
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid is not None",
                None,
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] is not None'.format(obj.prop),
            None,
        ]

    def _is_str_cond(self, obj):
        if cname(obj) in [
            "StringEq",
            "StringStartsWith",
            "StringEndsWith",
            "StringContains",
            "StringMatch",
        ]:
            return True
        return False

    def _is_num_cond(self, obj):
        if cname(obj) in [
            "NumEq",
            "NumLt",
            "NumGt",
            "NumLte",
            "NumGte",
            "NumNan",
            "NumNotNan",
            "NumInf",
            "NumNotInf",
        ]:
            return True
        return False

    def _parse_str(self, obj):
        if cname(obj) == "StringEq":
            return self._parse_str_eq(obj)
        if cname(obj) == "StringStartsWith":
            return self._parse_str_starts_with(obj)
        if cname(obj) == "StringEndsWith":
            return self._parse_str_ends_with(obj)
        if cname(obj) == "StringContains":
            return self._parse_str_contains(obj)
        if cname(obj) == "StringMatch":
            return self._parse_str_match(obj)
        raise RuntimeError("Bad String Op Class")

    def _parse_str_eq(self, obj):
        return [
            None,
            obj.name,
            'df_row["{}"] == "{}"'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], str)".format(obj.prop),
        ]

    def _parse_str_starts_with(self, obj):
        return [
            None,
            obj.name,
            'df_row["{}"].startswith("{}")'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], str)".format(obj.prop),
        ]

    def _parse_str_ends_with(self, obj):
        return [
            None,
            obj.name,
            'df_row["{}"].endswith("{}")'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], str)".format(obj.prop),
        ]

    def _parse_str_contains(self, obj):
        return [
            None,
            obj.name,
            '"{}" in df_row["{}"]'.format(obj.val, obj.prop),
            "isinstance(df_row['{}'], str)".format(obj.prop),
        ]

    def _parse_str_match(self, obj):
        return [
            None,
            obj.name,
            're.match("{}", df_row["{}"]) is not None'.format(obj.val, obj.prop),
            "isinstance(df_row['{}'], str)".format(obj.prop),
        ]

    def _parse_num(self, obj):
        if cname(obj) == "NumEq":
            return self._parse_num_eq(obj)
        if cname(obj) == "NumLt":
            return self._parse_num_lt(obj)
        if cname(obj) == "NumGt":
            return self._parse_num_gt(obj)
        if cname(obj) == "NumLte":
            return self._parse_num_lte(obj)
        if cname(obj) == "NumGte":
            return self._parse_num_gte(obj)
        if cname(obj) == "NumNan":
            return self._parse_num_nan(obj)
        if cname(obj) == "NumNotNan":
            return self._parse_num_not_nan(obj)
        if cname(obj) == "NumInf":
            return self._parse_num_inf(obj)
        if cname(obj) == "NumNotInf":
            return self._parse_num_not_inf(obj)
        raise RuntimeError("Bad Number Op Class")

    def _parse_num_eq(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "df_row.name._depth == {}".format(obj.val),
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid == {}".format(obj.val),
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] == {}'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_lt(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "df_row.name._depth < {}".format(obj.val),
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid < {}".format(obj.val),
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] < {}'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_gt(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "df_row.name._depth > {}".format(obj.val),
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid > {}".format(obj.val),
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] > {}'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_lte(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "df_row.name._depth <= {}".format(obj.val),
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid <= {}".format(obj.val),
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] <= {}'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_gte(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "df_row.name._depth >= {}".format(obj.val),
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid >= {}".format(obj.val),
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] >= {}'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_nan(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "pd.isna(df_row.name._depth)",
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "pd.isna(df_row.name._hatchet_nid)",
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'pd.isna(df_row["{}"])'.format(obj.prop),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_not_nan(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "not pd.isna(df_row.name._depth)",
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "not pd.isna(df_row.name._hatchet_nid)",
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'not pd.isna(df_row["{}"])'.format(obj.prop),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_inf(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "np.isinf(df_row.name._depth)",
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "np.isinf(df_row.name._hatchet_nid)",
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'np.isinf(df_row["{}"])'.format(obj.prop),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_not_inf(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "not np.isinf(df_row.name._depth)",
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "not np.isinf(df_row.name._hatchet_nid)",
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'not np.isinf(df_row["{}"])'.format(obj.prop),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]


class AndQuery(NaryQuery):
    """Compound Query that returns the intersection of the results
    of the subqueries"""

    def __init__(self, *args):
        """Create a new AndQuery object.

        Arguments:
            *args (tuple): the subqueries (high-level, low-level, or compound) to be performed.
        """
        if sys.version_info[0] == 2:
            super(AndQuery, self).__init__(args)
        else:
            super().__init__(args)
        if len(self.subqueries) < 2:
            raise BadNumberNaryQueryArgs("AndQuery requires 2 or more subqueries")

    def _perform_nary_op(self, query_results, gf):
        """Perform an intersection operation on the results of the subqueries.

        Arguments:
            query_results (list): the results of the subqueries.
            gf (GraphFrame): the GraphFrame on which the query is applied.

        Returns:
            (list): A list of nodes representing the intersection of the results of the subqueries.
        """
        intersection_set = set(query_results[0]).intersection(*query_results[1:])
        return list(intersection_set)


"""Alias of AndQuery to signify the relationship to set Intersection"""
IntersectionQuery = AndQuery


class OrQuery(NaryQuery):
    """Compound Query that returns the union of the results
    of the subqueries"""

    def __init__(self, *args):
        """Create a new OrQuery object.

        Arguments:
            *args (tuple): the subqueries (high-level, low-level, or compound) to be performed.
        """
        if sys.version_info[0] == 2:
            super(OrQuery, self).__init__(args)
        else:
            super().__init__(args)
        if len(self.subqueries) < 2:
            raise BadNumberNaryQueryArgs("OrQuery requires 2 or more subqueries")

    def _perform_nary_op(self, query_results, gf):
        """Perform an union operation on the results of the subqueries.

        Arguments:
            query_results (list): the results of the subqueries.
            gf (GraphFrame): the GraphFrame on which the query is applied.

        Returns:
            (list): A list of nodes representing the union of the results of the subqueries.
        """
        union_set = set().union(*query_results)
        return list(union_set)


"""Alias of OrQuery to signify the relationship to set Union"""
UnionQuery = OrQuery


class XorQuery(NaryQuery):
    """Compound Query that returns the symmetric difference
    (i.e., set-based XOR) of the results of the subqueries"""

    def __init__(self, *args):
        """Create a new XorQuery object.

        Arguments:
            *args (tuple): the subqueries (high-level, low-level, or compound) to be performed.
        """
        if sys.version_info[0] == 2:
            super(XorQuery, self).__init__(args)
        else:
            super().__init__(args)
        if len(self.subqueries) < 2:
            raise BadNumberNaryQueryArgs("XorQuery requires 2 or more subqueries")

    def _perform_nary_op(self, query_results, gf):
        """Perform a symmetric difference operation on the results of the subqueries.

        Arguments:
            query_results (list): the results of the subqueries.
            gf (GraphFrame): the GraphFrame on which the query is applied.

        Returns:
            (list): A list of nodes representing the symmetric difference of the results of the subqueries.
        """
        xor_set = set()
        for res in query_results:
            xor_set = xor_set.symmetric_difference(set(res))
        return list(xor_set)


"""Alias of XorQuery to signify the relationship to set Symmetric Difference"""
SymDifferenceQuery = XorQuery


class NotQuery(NaryQuery):
    """Compound Query that returns all nodes in the GraphFrame that
    are not returned from the subquery."""

    def __init__(self, *args):
        """Create a new XorQuery object.

        Arguments:
            *args (tuple): the subquery (high-level, low-level, or compound) to be performed.
        """
        if sys.version_info[0] == 2:
            super(NotQuery, self).__init__(args)
        else:
            super().__init__(args)
        if len(self.subqueries) != 1:
            raise BadNumberNaryQueryArgs("NotQuery requires exactly 1 subquery")

    def _perform_nary_op(self, query_results, gf):
        """Collect all nodes in the graph not present in the query result.

        Arguments:
            query_results (list): the result of the subquery.
            gf (GraphFrame): the GraphFrame on which the query is applied.

        Returns:
            (list): A list of all nodes not found in the subquery.
        """
        nodes = set(gf.graph.traverse())
        query_nodes = set(query_results[0])
        return list(nodes.difference(query_nodes))


class InvalidQueryPath(Exception):
    """Raised when a query does not have the correct syntax"""


class InvalidQueryFilter(Exception):
    """Raised when a query filter does not have a valid syntax"""


class BadNumberNaryQueryArgs(Exception):
    """Raised when a query filter does not have a valid syntax"""



# Copyright 2021-2024 University of Maryland and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd
import hatchet.graphframe
from hatchet.node import Node
from hatchet.graph import Graph
from hatchet.graphframe import GraphFrame
from hatchet.frame import Frame
import struct
import yaml


class MetaReader:
    # adds new context id and return new nid
    def _add_context_id(self, context_id) -> int:
        self.nid_to_ctx[self.current_nid] = context_id
        self.current_nid += 1
        return self.current_nid - 1

    def __init__(self, file_location):
        # open the file to ready in binary mode (rb)
        self.file = open(file_location, "rb")

        # setting necessary read options
        self.byte_order = "little"
        self.signed = False
        self.encoding = "ASCII"
        self.current_nid = 0
        self.nid_to_ctx = {}
        self.node_map = {}
        self.metric_id_name_map = {}
        self.number_of_metric_types = 0

        # The meta.db header consists of the common .db header and n sections.
        # We're going to do a little set up work, so that's easy to change if
        # any revisions change the orders.

        # We're are going to specify 2 maps:
        #   - One dictionary maps the section name to an index
        #       (which follows the order that the sections are seen
        #        in the meta.db header)
        #   - The second dictionary maps the section name to a
        #       function that reads the section. Each function is defined
        #       as __read_<section_name>_section(self, section_pointer: int,
        #       section_size: int) -> None

        # Here I'm mapping the section name to it's order in the meta.db header
        header_map = {
            "General Properties": 0,
            "Identifier Names": 1,
            "Performance Metrics": 2,
            "Context Tree": 3,
            "Common String Table": 4,
            "Load Modules": 5,
            "Source Files": 6,
            "Functions": 7,
        }

        # Now let's create a function to section map
        reader_map = {
            "General Properties": self.__read_general_properties_section,
            "Common String Table": self.__read_common_string_table_section,
            "Source Files": self.__read_source_files_section,
            "Functions": self.__read_functions_section,
            "Load Modules": self.__read_load_modules_section,
            "Context Tree": self.__read_context_tree_section,
            "Identifier Names": self.__read_identifier_names_section,
            "Performance Metrics": self.__read_performance_metrics_section,
        }

        # Another thing thing that we should consider is the order to read the sections.
        # Here is a list of section references (x -> y means x references y)
        #   - "Source Files"    ->  "Common String Table"
        #   - "Functions"       ->  "Common String Table"
        #   - "Context Tree"    ->  "Common String Table"
        #   - "Load Modules"    ->  "Common String Table"
        #   - "Functions"       ->  "Source Files"
        #   - "Context Tree"    ->  "Functions"
        #   - "Context Tree"    ->  "Source Files"
        #   - "Functions"       ->  "Load Modules"
        #
        # Thus we want to keep this order when reading sections:
        # "Common String Table" -> "Source Files" -> "Functions" -> "Context Tree", and
        # "Common String Table -> "Load Modules" -> "Context Tree"
        # Here I'm specifying the order of reading the file
        self.read_order = [
            "Common String Table",
            "General Properties",
            "Source Files",
            "Load Modules",
            "Functions",
            "Context Tree",
            "Identifier Names",
            "Performance Metrics",
        ]

        # Let's make sure that we include every section in the read order and reader_map
        assert set(self.read_order) == set(header_map) and set(header_map) == set(
            reader_map
        )

        # Now to the actual reading of the meta.db file

        # reading the meta.db header
        self.__read_common_header()

        # now let's read all the sections
        for section_name in self.read_order:
            section_index = header_map[section_name]
            section_pointer = self.section_pointer[section_index]
            section_size = self.section_size[section_index]
            section_reader = reader_map[section_name]
            section_reader(section_pointer, section_size)

    def get_information_from_context_id(self, context_id: int):
        context: dict = self.context_map[context_id]
        if "string_index" in context:
            return {
                "module": "",
                "file": "",
                "function": self.common_strings[context["string_index"]],
                "relation": -1,
                "lexical_type": -1,
                "line": -1,
                "loop_type": False,
            }

        load_module_index = context["load_module_index"]
        source_file_index = context["source_file_index"]
        source_file_line = context["source_file_line"]
        function_index = context["function_index"]
        lexical_type = context["lexical_type"]
        load_module_offset = context["load_module_offset"]

        source_file_string = None
        module_string = None
        function_string = None
        file_line = -1
        type = None

        if lexical_type == 1:
            # loop construct
            function_string = "loop"
            type = "loop"
        elif lexical_type == 2:
            type = "line"
        elif lexical_type == 3:
            type = "instruction"
        elif lexical_type == 0:
            type = "function"
        else:
            # function is unkown
            function_string = "<unkown function>"

        if load_module_offset is not None:
            load_module_offset = hex(int(load_module_offset))

        if load_module_index is not None:
            load_module = self.load_modules_list[load_module_index]
            module_string = self.common_strings[load_module["string_index"]]
        if source_file_index is not None:
            source_file = self.source_files_list[source_file_index]
            source_file_string = self.common_strings[source_file["string_index"]]
        if function_index is not None:
            function = self.functions_list[function_index]
            function_string = self.common_strings[function["string_index"]]
        if source_file_line is not None:
            file_line = str(source_file_line)
        return {
            "module": module_string,
            "file": source_file_string,
            "function": function_string,
            "line": file_line,
            "type": type,
            "relation": context["relation"],
            "instruction": load_module_offset,
            # "loop_type": loop_type,
        }

    def __read_common_header(self) -> None:
        """
        Reads common .db file header version 4.0
        """

        # read Magic identifier ("HPCPROF-tracedb_")
        # first ten buyes are HPCTOOLKIT in ASCII
        identifier = str(self.file.read(10), encoding=self.encoding)
        assert identifier == "HPCTOOLKIT"

        # next 4 bytes (u8) are the "Specific format identifier"
        format_identifier = str(self.file.read(4), encoding=self.encoding)
        assert format_identifier == "meta"

        # next byte (u8) contains the "Common major version, currently 4"
        self.major_version = int.from_bytes(
            self.file.read(1), byteorder=self.byte_order, signed=self.signed
        )
        # next byte (u8) contains the "Specific minor version"
        self.minor_version = int.from_bytes(
            self.file.read(1), byteorder=self.byte_order, signed=self.signed
        )

        self.section_pointer = []
        self.section_size = []
        # In the header each section is given 16 bytes:
        #   - First 8 bytes specify the total size of the section (in bytes)
        #   - Last 8 bytes specify a pointer to the beggining of the section
        for i in range(len(self.read_order)):
            self.section_size.append(
                int.from_bytes(
                    self.file.read(8), byteorder=self.byte_order, signed=self.signed
                )
            )
            self.section_pointer.append(
                int.from_bytes(
                    self.file.read(8), byteorder=self.byte_order, signed=self.signed
                )
            )

    def __read_general_properties_section(
        self, section_pointer: int, section_size: int
    ) -> None:
        """Reads the general properties of the trace.
        Sets:

        self.database_title: Title of the database. May be provided by the user.

        self.database_description: Human-readable Markdown description of the database.
        """

        # go to the right spot in the file
        self.file.seek(section_pointer)
        title_pointer = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )
        description_pointer = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )

        self.database_title = self.__read_string(title_pointer)
        self.database_description = self.__read_string(description_pointer)

    def __get_common_string(self, string_pointer: int) -> str:
        """Given the file pointer to find string, returns the string."""
        if string_pointer in self.common_string_index_map:
            return self.common_strings[self.common_string_index_map[string_pointer]]
        else:
            return None

    def __read_common_string_table_section(
        self, section_pointer: int, section_size: int
    ) -> None:
        # Let's go to the section
        self.file.seek(section_pointer)

        # We know that this section is just a densely packed list of strings,
        # seperated by the null character
        # So to create a list of these strings, we'll read them all into one string then
        # split them by the null character

        # Reading entire section into a string
        total_section: str = str(self.file.read(section_size), encoding="UTF-8")

        # Splitting entire section into list of strings
        self.common_strings: list[str] = total_section.split("\0")

        # Now we are creating a map between the original location to the string
        # to the index of the string in self.common_strings.
        # This is because we are passed pointers to find the string in other sections
        pointer_index = section_pointer
        # pointer_index = 0
        self.common_string_index_map: dict = {}
        for i in range(len(self.common_strings)):
            self.common_string_index_map[pointer_index] = i
            pointer_index += len(self.common_strings[i]) + 1

    def __get_load_modules_index(self, load_module_pointer: int) -> int:
        """
        Given the pointer to where the file would exists in meta.db,
        returns the index of the file in self.source_files_list.
        """
        return (
            load_module_pointer - self.load_modules_pointer
        ) // self.load_module_size

    def __read_load_modules_section(
        self, section_pointer: int, section_size: int
    ) -> None:
        """
        Reads the "Load Modules" Section of meta.db.
        """
        # go to the right spot in meta.db
        self.file.seek(section_pointer)

        # Load modules used in this database
        self.load_modules_pointer = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )
        # Number of load modules listed in this section (u32)
        num_load_modules = int.from_bytes(
            self.file.read(4), byteorder=self.byte_order, signed=self.signed
        )
        # Size of a Load Module Specification, currently 16 (u16)
        self.load_module_size = int.from_bytes(
            self.file.read(2), byteorder=self.byte_order, signed=self.signed
        )

        # Going to store file's path in self.load_modules_list.
        # Each will contain the index of file's path string in
        # self.common_string
        self.load_modules_list: list[dict] = []

        for i in range(num_load_modules):
            current_index = self.load_modules_pointer + (i * self.load_module_size)
            self.file.seek(current_index)

            # Flags -- Reserved for future use (u32)
            # flags = int.from_bytes(
            #     self.file.read(4), byteorder=self.byte_order, signed=self.signed
            # )
            self.file.read(4)
            # empty space that we need to skip
            self.file.read(4)
            # Full path to the associated application binary
            path_pointer = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )
            module_map = {"string_index": self.common_string_index_map[path_pointer]}
            self.load_modules_list.append(module_map)

    def __read_string(self, file_pointer: int) -> str:
        """
        Helper function to read a string from the file starting at the file_pointer
        and ending at the first occurence of the null character
        """
        self.file.seek(file_pointer)
        name = ""
        while True:
            read = str(self.file.read(1), encoding="UTF-8")
            if read == "\0":
                break
            name += read
        return name

    def get_identifier_name(self, kind: int):
        """
        returns the identifier name, given the kind
        """
        return self.identifier_names[kind]

    def __read_identifier_names_section(
        self, section_pointer: int, section_size: int
    ) -> None:
        """
        Reads "Identifier Names" Section and Identifier Name strings in self.names_list
        """

        # go to correct section of file
        self.file.seek(section_pointer)

        # Human-readable names for Identifier kinds
        names_pointer_pointer = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )
        # Number of names listed in this section
        num_names = int.from_bytes(
            self.file.read(1), byteorder=self.byte_order, signed=self.signed
        )

        self.identifier_names: list[str] = []

        for i in range(num_names):
            self.file.seek(names_pointer_pointer + (i * 8))
            names_pointer = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )
            identifier_name = self.__read_string(names_pointer)
            if identifier_name == "NODE":
                self.identifier_names.append("node_pid")
            else:
                self.identifier_names.append(identifier_name.lower())

    def __read_performance_metrics_section(
        self, section_pointer: int, section_size: int
    ) -> None:
        # go to correct section in file
        self.file.seek(section_pointer)

        # {MD} Descriptions of performance metrics
        md_pointer = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )

        number_of_md = int.from_bytes(
            self.file.read(4), byteorder=self.byte_order, signed=self.signed
        )

        # Size of the {MD} structure, currently 32
        size_of_md = int.from_bytes(
            self.file.read(1), byteorder=self.byte_order, signed=self.signed
        )

        # Size of the {PSI} structure, currently 16
        size_of_psi = int.from_bytes(
            self.file.read(1), byteorder=self.byte_order, signed=self.signed
        )

        # Size of the {SS} structure, currently 24
        self.file.read(1)
        # size_of_ss = int.from_bytes(
        #     self.file.read(1), byteorder=self.byte_order, signed=self.signed
        # )

        # empty space
        self.file.read(1)

        # {PS} Descriptions of propgation scopes.
        # We don't need the descriptions at this point.
        # We will read descriptions for each metric later.
        self.file.read(8)
        # ps_pointer = int.from_bytes(
        #     self.file.read(8), byteorder=self.byte_order, signed=self.signed
        # )

        # Number of propgation scopes
        self.file.read(2)
        # number_of_ps = int.from_bytes(
        #     self.file.read(2), byteorder=self.byte_order, signed=self.signed
        # )

        # Size of the {PS} structure, currently 16
        self.file.read(1)
        # size_of_ps = int.from_bytes(
        #     self.file.read(1), byteorder=self.byte_order, signed=self.signed
        # )

        # Read MD
        for i in range(number_of_md):
            self.file.seek(md_pointer + (size_of_md * i))
            metric_name_pointer = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )

            # {PSI} Instantiated propagated sub-metrics
            psi_pointer = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )

            # {SS} Summary statistics. Hatchet currently don't need summary statistics.
            self.file.read(8)
            # ss_pointer = int.from_bytes(
            #     self.file.read(8), byteorder=self.byte_order, signed=self.signed
            # )

            # Number of instantiated sub-metrics for this metric
            self.number_of_metric_types = int.from_bytes(
                self.file.read(2), byteorder=self.byte_order, signed=self.signed
            )

            # Number of summary statistics for this metric. Hatchet currently
            # don't need summary statistics.
            self.file.read(2)
            # number_of_ss = int.from_bytes(
            #     self.file.read(2), byteorder=self.byte_order, signed=self.signed
            # )

            # empty space
            self.file.read(4)

            metric_name = self.__read_string(metric_name_pointer)

            # Instantiated propagated sub-metrics (PSI)
            # Copied from FORMATS.md: Propagated metric values are generated for each
            # context by summing values attributed to children contexts, within the
            # measurements for a single application thread. Which children are included
            # in this sum is indicated by the *pScope PS structure.
            for i in range(self.number_of_metric_types):
                self.file.seek(psi_pointer + (size_of_psi * i))

                ps_in_psi_pointer = int.from_bytes(
                    self.file.read(8), byteorder=self.byte_order, signed=self.signed
                )

                psi_metric_id = int.from_bytes(
                    self.file.read(2), byteorder=self.byte_order, signed=self.signed
                )

                # empty space
                self.file.read(6)

                # Get the description of propagation scopes for this metric ID.
                self.file.seek(ps_in_psi_pointer)

                # Examples: point, function, lex_aware, execution
                psi_name_pointer = int.from_bytes(
                    self.file.read(8), byteorder=self.byte_order, signed=self.signed
                )

                # Differet types:
                # 1: point, 2: execution, 3: function, 4: lex-aware.
                self.file.read(1)
                # psi_type = int.from_bytes(
                #     self.file.read(1), byteorder=self.byte_order, signed=self.signed
                # )

                self.file.read(1)
                # psi_index = int.from_bytes(
                #     self.file.read(1), byteorder=self.byte_order, signed=self.signed
                # )

                # empty space
                self.file.read(6)

                psi_name = self.__read_string(psi_name_pointer)
                self.metric_id_name_map[psi_metric_id] = (metric_name, psi_name)

    def __get_function_index(self, function_pointer: int) -> int:
        """
        Given the pointer to where the function would exists in meta.db,
        returns the index of the file in self.functions_list.
        """
        index = (function_pointer - self.functions_array_pointer) // self.function_size
        assert index < len(self.functions_list)
        return index

    def __read_functions_section(self, section_pointer: int, section_size: int) -> None:
        """
        Reads the "Functions" section of meta.db.
        """

        # go to correct section in file
        self.file.seek(section_pointer)

        self.functions_array_pointer = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )
        num_functions = int.from_bytes(
            self.file.read(4), byteorder=self.byte_order, signed=self.signed
        )
        self.function_size = int.from_bytes(
            self.file.read(2), byteorder=self.byte_order, signed=self.signed
        )

        self.functions_list: list[dict] = []
        for i in range(num_functions):
            current_index = self.functions_array_pointer + (i * self.function_size)
            self.file.seek(current_index)
            function_name_pointer = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )
            modules_pointer = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )
            modules_offset = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )
            file_pointer = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )
            source_line = int.from_bytes(
                self.file.read(4), byteorder=self.byte_order, signed=self.signed
            )
            # flags = int.from_bytes(
            #     self.file.read(4), byteorder=self.byte_order, signed=self.signed
            # )
            self.file.read(4)
            source_file_index = None
            load_module_index = None
            function_name_index = None
            if function_name_pointer != 0:
                function_name_index = self.common_string_index_map[
                    function_name_pointer
                ]
            if modules_pointer != 0:
                load_module_index = self.__get_load_modules_index(modules_pointer)
                # currently ignoring offset -- no idea how that's used
            if file_pointer != 0:
                source_file_index = self.__get_source_file_index(file_pointer)

            current_function_map = {
                "string_index": function_name_index,
                "source_line": source_line,
                "load_modules_index": load_module_index,
                "source_file_index": source_file_index,
                "load_modules_offset": modules_offset,
            }
            self.functions_list.append(current_function_map)

    def __get_source_file_index(self, source_file_pointer: int) -> int:
        """
        Given the pointer to where the file would exists in meta.db,
        returns the index of the file in self.source_files_list.
        """
        index = (
            source_file_pointer - self.source_files_pointer
        ) // self.source_file_size
        assert index < len(self.source_files_list)
        return index

    def __read_source_files_section(
        self, section_pointer: int, section_size: int
    ) -> None:
        """
        Reads the "Source Files" Section of meta.db.
        """

        self.file.seek(section_pointer)

        # Source files used in this database
        self.source_files_pointer = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )

        # Number of source files listed in this section (u32)
        num_files = int.from_bytes(
            self.file.read(4), byteorder=self.byte_order, signed=self.signed
        )

        # Size of a Source File Specification, currently 16 (u16)
        self.source_file_size = int.from_bytes(
            self.file.read(2), byteorder=self.byte_order, signed=self.signed
        )

        # Looping through individual files to get there information now
        self.file.seek(self.source_files_pointer)

        # Going to store file's path in self.files_list.
        # Each will contain the index of file's path string in
        # self.common_string
        self.source_files_list: list[dict] = []
        for i in range(num_files):
            # Reading information about each individual source file
            self.file.seek(self.source_files_pointer + (i * self.source_file_size))

            # flag = int.from_bytes(
            #     self.file.read(4), byteorder=self.byte_order, signed=self.signed
            # )
            self.file.read(4)
            # empty space that we need to skip
            self.file.read(4)
            # Path to the source file. Absolute, or relative to the root database
            # directory. The string pointed to by pPath is completely within the
            # Common String Table section, including the terminating NUL byte.
            file_path_pointer = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )
            string_index = self.common_string_index_map[file_path_pointer]
            source_file_map = {"string_index": string_index}
            self.source_files_list.append(source_file_map)

    def __read_context_tree_section(
        self, section_pointer: int, section_size: int
    ) -> None:
        """
        Reads the "Context Tree" section of meta.db.

        Loops and calls __read_single_entry_point with the correct pointer to read
        the correct entry and add it to the CCT.
        """

        self.roots = []
        self.context_map: dict[int, dict] = {}

        # Reading "Context Tree" section header

        # make sure we're in the right spot of the file
        self.file.seek(section_pointer)

        # ({Entry}[nEntryPoints]*)
        entry_points_array_pointer = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )
        # (u16)
        num_entry_points = int.from_bytes(
            self.file.read(2), byteorder=self.byte_order, signed=self.signed
        )
        # (u8)
        entry_point_size = int.from_bytes(
            self.file.read(1), byteorder=self.byte_order, signed=self.signed
        )

        for i in range(num_entry_points):
            current_pointer = entry_points_array_pointer + (i * entry_point_size)
            self.__read_single_entry_point(current_pointer)

    def __read_single_entry_point(self, entry_point_pointer: int) -> None:
        """
        Reads single (root) context entry.

        Reads the correct entry and adds it to the CCT.
        """

        self.file.seek(entry_point_pointer)

        # Reading information about child contexts
        # Total size of *pChildren (I call pChildren children_pointer), in bytes (u64)
        children_size = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )
        # Pointer to the array of child contexts
        children_pointer = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )

        # Reading information about this context
        # Unique identifier for this context (u32)
        context_id = int.from_bytes(
            self.file.read(4), byteorder=self.byte_order, signed=self.signed
        )
        # Type of entry point used here (u16)
        # entry_point_type = int.from_bytes(
        #     self.file.read(2), byteorder=self.byte_order, signed=self.signed
        # )
        self.file.read(2)
        # next 2 bytes are blank
        self.file.read(2)
        # Human-readable name for the entry point
        pretty_name_pointer = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )

        # map context for this context
        string_index = self.common_string_index_map[pretty_name_pointer]

        context = {
            "relation": -1,
            "lexical_type": -1,
            "function_index": -1,
            "source_file_index": -1,
            "source_file_line": -1,
            "load_module_index": -1,
            "load_module_offset": -1,
            "string_index": string_index,
        }
        self.context_map[context_id] = context
        # Create Node for this context
        root_context_info = self.get_information_from_context_id(context_id)
        node = Node(
            Frame({"type": "function", "name": root_context_info["function"]}),
            parent=None,
            hnid=self._add_context_id(context_id),
        )

        # Adding the Node to the CCT
        self.roots.append(node)
        self.node_map[context_id] = node

        # Reading the children contexts
        self.__read_children_contexts(children_pointer, children_size, node, context_id)

    def __read_children_contexts(
        self,
        context_array_pointer: int,
        total_size: int,
        parent_node: Node,
        parent_context_id: int,
    ) -> None:
        """
        Recursive function to read all child contexts and add it to the CCT
        """

        if total_size <= 0 or context_array_pointer <= 0:
            return
        self.file.seek(context_array_pointer)
        index = 0
        while index < total_size:
            # Reading information about child contexts (as in the children of
            # this context)
            # Total size of *pChildren (I call pChildren children_pointer),
            # in bytes (u64)
            children_size = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )
            index += 8
            # Pointer to the array of child contexts
            children_pointer = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )
            index += 8

            # Reading information about this context
            # Unique identifier for this context (u32)
            context_id = int.from_bytes(
                self.file.read(4), byteorder=self.byte_order, signed=self.signed
            )
            index += 4
            # Reading flags (u8)
            flags = int.from_bytes(
                self.file.read(1), byteorder=self.byte_order, signed=self.signed
            )
            index += 1
            # Relation this context has with its parent (u8)
            relation = int.from_bytes(
                self.file.read(1), byteorder=self.byte_order, signed=self.signed
            )
            index += 1
            # Type of lexical context represented (u8)
            lexical_type = int.from_bytes(
                self.file.read(1), byteorder=self.byte_order, signed=self.signed
            )
            index += 1
            # Size of flex, in u8[8] "words" (bytes / 8) (u8)
            num_flex_words = int.from_bytes(
                self.file.read(1), byteorder=self.byte_order, signed=self.signed
            )
            index += 1
            # Bitmask for defining propagation scopes (u16)
            # propogation = int.from_bytes(
            #     self.file.read(2), byteorder=self.byte_order, signed=self.signed
            # )
            self.file.read(2)
            index += 2
            # Empty space
            self.file.read(6)
            index += 6

            # reading flex
            flex = self.file.read(8 * num_flex_words)
            index += 8 * num_flex_words

            function_index: int = None
            source_file_index: int = None
            source_file_line: int = None
            load_module_index: int = None
            load_module_offset: int = None

            # flex is u8[8][num_flex],
            # meaning that one flex word is 8 bytes or u64

            # Bit 0: hasFunction. If 1, the following sub-fields of flex are present:
            #   flex[0]: FS* pFunction: Function associated with this context
            if flags & 1 != 0:
                sub_flex = int.from_bytes(
                    flex[0:8], byteorder=self.byte_order, signed=self.signed
                )
                flex = flex[8:]
                function_index = self.__get_function_index(sub_flex)

            # Bit 1: hasSrcLoc. If 1, the following sub-fields of flex are present:
            #   flex[1]: SFS* pFile: Source file associated with this context
            #   flex[2]: u32 line: Associated source line in pFile
            if flags & 2 != 0:
                sub_flex_1 = int.from_bytes(
                    flex[0:8], byteorder=self.byte_order, signed=self.signed
                )
                sub_flex_2 = int.from_bytes(
                    flex[8:10], byteorder=self.byte_order, signed=self.signed
                )
                flex = flex[16:]
                source_file_index = self.__get_source_file_index(sub_flex_1)
                source_file_line = sub_flex_2

            # Bit 2: hasPoint. If 1, the following sub-fields of flex are present:
            #   flex[3]: LMS* pModule: Load module associated with this context
            #   flex[4]: u64 offset: Associated byte offset in *pModule
            if flags & 4 != 0:
                sub_flex_1 = int.from_bytes(
                    flex[0:8], byteorder=self.byte_order, signed=self.signed
                )
                sub_flex_2 = int.from_bytes(
                    flex[8:16], byteorder=self.byte_order, signed=self.signed
                )
                flex = flex[16:]
                load_module_index = self.__get_load_modules_index(sub_flex_1)
                load_module_offset = sub_flex_2

            if lexical_type == 0:
                # function call
                # this means that information about the
                # source file and module are with the parent
                parent_information = self.context_map[parent_context_id]
                if "string_index" in parent_information and function_index is not None:
                    # This means that the parent is the root, and it's
                    # information is useless
                    function = self.functions_list[function_index]
                    # Getting source file line
                    source_file_line = function["source_line"]
                    # Getting source file name
                    source_file_index = function["source_file_index"]
                    # Getting Load Module Name
                    load_module_index = function["load_modules_index"]
                    # Getting Load Module Offset

                    load_module_offset = function["load_modules_offset"]
                else:
                    if source_file_index is None:
                        source_file_index = parent_information["source_file_index"]
                    if source_file_line is None:
                        source_file_line = parent_information["source_file_line"]
                    if load_module_index is None:
                        load_module_index = parent_information["load_module_index"]
                        load_module_offset = parent_information["load_module_offset"]

            # creating a map for this context
            context = {
                "relation": relation,
                "lexical_type": lexical_type,
                "function_index": function_index,
                "source_file_index": source_file_index,
                "source_file_line": source_file_line,
                "load_module_index": load_module_index,
                "load_module_offset": load_module_offset,
            }

            self.context_map[context_id] = context
            node_context_info = self.get_information_from_context_id(context_id)
            node = None
            if node_context_info["type"] == "line":
                node = Node(
                    Frame(
                        {
                            "type": "line",
                            "file": node_context_info["file"],
                            "line": node_context_info["line"],
                        }
                    ),
                    parent=parent_node,
                    hnid=self._add_context_id(context_id),
                )
            elif node_context_info["type"] == "instruction":
                node = Node(
                    Frame(
                        {
                            "type": "instruction",
                            "module": node_context_info["module"],
                            "instruction": node_context_info["instruction"],
                        }
                    ),
                    parent=parent_node,
                    hnid=self._add_context_id(context_id),
                )
            elif node_context_info["type"] == "loop":
                node = Node(
                    Frame(
                        {
                            "type": "loop",
                            "file": node_context_info["file"],
                            "line": node_context_info["line"],
                        }
                    ),
                    parent=parent_node,
                    hnid=self._add_context_id(context_id),
                )
            else:
                node = Node(
                    Frame({"type": "function", "name": node_context_info["function"]}),
                    parent=parent_node,
                    hnid=self._add_context_id(context_id),
                )

            # Connecting this node to the parent node
            parent_node.add_child(node)
            next_parent_node = node

            # Adding this node to the graph
            self.node_map[context_id] = node

            # recursively call this function to add more children
            return_address = self.file.tell()
            self.__read_children_contexts(
                children_pointer, children_size, next_parent_node, context_id
            )
            self.file.seek(return_address)


class ProfileReader:
    # class to read self.data from profile.db file

    def __init__(self, file_location, meta_reader):
        # gets the pi_ptr variable to be able to read the identifier tuples
        self.meta_reader: MetaReader = meta_reader

        # open the file to ready in binary mode (rb)
        self.file = open(file_location, "rb")

        # setting necessary read options
        self.byte_order = "little"
        self.signed = False
        self.encoding = "ASCII"

        # The profile.db header consists of the common .db header and n sections.
        # We're going to do a little set up work, so that's easy to change if
        # any revisions change the orders.

        # We're are going to specify 2 maps:
        #   - One dictionary maps the section name to an index
        #       (which follows the order that the sections are seen
        #        in the meta.db header)
        #   - The second dictionary maps the section name to a
        #       function that reads the section. Each function is defined
        #       as __read_<section_name>_section(self, section_pointer: int,
        #       section_size: int) -> None

        # Here I'm mapping the section name to it's order in the meta.db header
        header_map = {"Profiles Information": 0, "Hierarchical Identifier Tuples": 1}

        # Now let's create a function to section map
        reader_map = {
            "Profiles Information": self.__read_profiles_information_section,
            "Hierarchical Identifier Tuples": self.__read_hit_section,
        }

        # Another thing thing that we should consider is the order to read the sections.
        # Here is a list of section references (x -> y means x references y)
        #   - "Profiles Information" -> "Hierarchical Identifier Tuples"
        #
        # Thus we want to keep this order when reading sections:
        # "Profiles Information" -> "Hierarchical Identifier Tuples"
        # Here I'm specifying the order of reading the file
        self.read_order = ["Hierarchical Identifier Tuples", "Profiles Information"]

        # Let's make sure that we include every section in the read order and reader_map
        assert set(self.read_order) == set(header_map) and set(header_map) == set(
            reader_map
        )

        # Now to the actual reading of the meta.db file

        # reading the meta.db header
        self.__read_common_header()

        # now let's read all the sections
        for section_name in self.read_order:
            section_index = header_map[section_name]
            section_pointer = self.section_pointer[section_index]
            section_size = self.section_size[section_index]
            section_reader = reader_map[section_name]
            section_reader(section_pointer, section_size)

    def __read_profiles_information_section(
        self, section_pointer: int, section_size: int
    ) -> None:
        """
        Reads Profile Information section.
        """

        self.file.seek(section_pointer)

        # Description for each profile (u64)
        profiles_pointer = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )
        # Number of profiles listed in this section (u32)
        num_profiles = int.from_bytes(
            self.file.read(4), byteorder=self.byte_order, signed=self.signed
        )
        # Size of a {PI} structure, currently 40 (u8)
        profile_size = int.from_bytes(
            self.file.read(1), byteorder=self.byte_order, signed=self.signed
        )

        self.profile_info_list = []
        for i in range(num_profiles):
            file_index = profiles_pointer + (i * profile_size)
            self.file.seek(file_index)

            # Profile-Major Sparse Value Block.
            # We don't need to read this section unless we want
            # the summary statistics. The commented out lines below
            # show how to read PSVB in case we read it in future.
            self.file.read(0x20)
            # psvb = self.file.read(0x20)

            # Identifier tuple for this profile
            hit_pointer = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )
            # (u32)
            flags = int.from_bytes(
                self.file.read(4), byteorder=self.byte_order, signed=self.signed
            )
            profile_map = {"hit_pointer": hit_pointer, "flags": flags}

            self.profile_info_list.append(profile_map)
            if hit_pointer == 0:
                # this is a summary profile
                self.summary_profile_index = i

    def get_hit_from_profile(self, index: int) -> list:
        profile = self.profile_info_list[index]
        hit_pointer = profile["hit_pointer"]
        if hit_pointer != 0:
            return self.hit_map[hit_pointer]
        else:
            profile = self.profile_info_list[self.summary_profile_index]
            hit_pointer = profile["hit_pointer"]
            return self.hit_map[hit_pointer]

    def __read_hit_section(self, section_pointer: int, section_size: int) -> None:
        """
        Reads Hierarchical Identifier Tuples section of profile.db
        """
        # let's get to the correct spot in the file
        self.file.seek(section_pointer)

        self.hit_map = {}

        while (self.file.tell() - section_pointer) < section_size:
            # hit pointer
            hit_pointer = self.file.tell()

            # Number of identifications in this tuple (u16)
            num_tuples = int.from_bytes(
                self.file.read(2), byteorder=self.byte_order, signed=self.signed
            )

            # empty space
            self.file.read(6)

            # Identifications for an application thread
            # Read H.I.T.s
            profile_properties = {}
            for i in range(num_tuples):
                # One of the values listed in the meta.db Identifier Names section. (u8)
                kind = int.from_bytes(
                    self.file.read(1), byteorder=self.byte_order, signed=self.signed
                )
                # empty space
                self.file.read(1)

                # flag
                flags = int.from_bytes(
                    self.file.read(2), byteorder=self.byte_order, signed=self.signed
                )
                # self.file.read(2)

                # Logical identifier value, may be arbitrary but dense towards 0. (u32)
                logical_id = int.from_bytes(
                    self.file.read(4), byteorder=self.byte_order, signed=self.signed
                )
                # self.file.read(4)

                # Physical identifier value, eg. hostid or PCI bus index. (u64)
                physical_id = int.from_bytes(
                    self.file.read(8), byteorder=self.byte_order, signed=self.signed
                )

                identifier_name = self.meta_reader.get_identifier_name(kind)

                if flags & 0x1:
                    profile_properties[identifier_name] = physical_id
                else:
                    profile_properties[identifier_name] = logical_id
                # tuples_list.append((identifier_name, physical_id, flags, logical_id))
            self.hit_map[hit_pointer] = profile_properties

    def __read_common_header(self) -> None:
        """
        Reads common .db file header version 4.0
        """

        # read Magic identifier ("HPCPROF-tracedb_")
        # first ten buyes are HPCTOOLKIT in ASCII
        identifier = str(self.file.read(10), encoding=self.encoding)
        assert identifier == "HPCTOOLKIT"

        # next 4 bytes (u8) are the "Specific format identifier"
        format_identifier = str(self.file.read(4), encoding=self.encoding)
        assert format_identifier == "prof"

        # next byte (u8) contains the "Common major version, currently 4"
        self.major_version = int.from_bytes(
            self.file.read(1), byteorder=self.byte_order, signed=self.signed
        )
        # next byte (u8) contains the "Specific minor version"
        self.minor_version = int.from_bytes(
            self.file.read(1), byteorder=self.byte_order, signed=self.signed
        )

        self.section_pointer = []
        self.section_size = []
        # In the header each section is given 16 bytes:
        #   - First 8 bytes specify the total size of the section (in bytes)
        #   - Last 8 bytes specify a pointer to the beggining of the section
        for i in range(len(self.read_order)):
            self.section_size.append(
                int.from_bytes(
                    self.file.read(8), byteorder=self.byte_order, signed=self.signed
                )
            )
            self.section_pointer.append(
                int.from_bytes(
                    self.file.read(8), byteorder=self.byte_order, signed=self.signed
                )
            )


class DefaultReader:
    """Reads the defaults.yaml file in the metrics directory.
    This file is needed to obtain inclusive and exclusive metrics."""

    def __init__(self, file_location: str, meta_reader: MetaReader) -> None:
        self.meta_reader = meta_reader
        self.metric_defaults = {}

        with open(file_location, "rb") as self.file:
            file_contents = yaml.safe_load(self.file)

            # Gets the inclusive and exclusive types for each metric.
            # This part might need modifications in future.
            # Reads the metric information in default.yaml and gets the
            # inclusive and exclusive metrics by looking at corresponding
            # section (inclusive/standard and exlusive/standard) of each
            # metric. Inclusive metric is usually "execution",
            # exclusive metric is usually "function".
            for metric in file_contents["roots"]:
                self.metric_defaults[metric["name"]] = {}
                if "Sum" in metric["variants"]:
                    if "percent" in metric["variants"]["Sum"]["render"]:
                        formula = metric["variants"]["Sum"]["formula"]
                        if "inclusive" in formula:
                            self.metric_defaults[metric["name"]]["inclusive"] = formula[
                                "inclusive"
                            ]["standard"]["scope"]

                        if "exclusive" in formula:
                            self.metric_defaults[metric["name"]]["exclusive"] = formula[
                                "exclusive"
                            ]["standard"]["scope"]

                # TODO: Not sure how to get inclusive and exclusive metrics when
                # we have a list in the standard section instead of a single variable.
                elif "" in metric["variants"]:
                    if "percent" in metric["variants"][""]["render"]:
                        formula = metric["variants"][""]["formula"]
                        if "inclusive" in formula:
                            self.metric_defaults[metric["name"]]["inclusive"] = formula[
                                "inclusive"
                            ]["standard"]["scope"]

                        if "exclusive" in formula:
                            self.metric_defaults[metric["name"]]["exclusive"] = formula[
                                "exclusive"
                            ]["standard"]["scope"]


class CCTReader:
    """Reads the metric and profile information (thread, rank, etc.)
    for each context."""

    def __init__(
        self,
        file_location: str,
        meta_reader: MetaReader,
        profile_reader: ProfileReader,
        defaults_reader: DefaultReader,
    ) -> None:
        # open file
        self.file = open(file_location, "rb")
        self.meta_reader = meta_reader
        self.profile_reader = profile_reader
        self.defaults_reader = defaults_reader
        self.node_dicts = {}
        self.inclusive_metrics = set([])
        self.exclusive_metrics = set([])

        # setting necessary read options
        self.byte_order = "little"
        self.signed = False
        self.encoding = "ASCII"

        # The cct.db header consists of the common .db header and n sections.
        # We're going to do a little set up work, so that's easy to change if
        # any revisions change the orders.

        # We're are going to specify 2 maps:
        #   - One dictionary maps the section name to an index
        #       (which follows the order that the sections are seen
        #        in the meta.db header)
        #   - The second dictionary maps the section name to a
        #       function that reads the section. Each function is defined
        #       as __read_<section_name>_section(self, section_pointer: int,
        #       section_size: int) -> None

        # Here I'm mapping the section name to it's order in the meta.db header
        header_map = {"Context Headers": 0}

        # Now let's create a function to section map
        reader_map = {"Context Headers": self.__read_cct_info_section}

        # Another thing thing that we should consider is the order to read the sections.
        # Here I'm specifying the order of reading the file
        self.read_order = ["Context Headers"]

        # Let's make sure that we include every section in the read order and reader_map
        assert set(self.read_order) == set(header_map) and set(header_map) == set(
            reader_map
        )

        # Now to the actual reading of the meta.db file

        # reading the meta.db header
        self.__read_common_header()

        # now let's read all the sections
        for section_name in self.read_order:
            section_index = header_map[section_name]
            section_pointer = self.section_pointer[section_index]
            section_size = self.section_size[section_index]
            section_reader = reader_map[section_name]
            section_reader(section_pointer, section_size)

    def __read_common_header(self) -> None:
        """
        Reads common .db file header version 4.0
        """

        # read Magic identifier ("HPCPROF-tracedb_")
        # first ten buyes are HPCTOOLKIT in ASCII
        identifier = str(self.file.read(10), encoding=self.encoding)
        assert identifier == "HPCTOOLKIT"

        # next 4 bytes (u8) are the "Specific format identifier"
        format_identifier = str(self.file.read(4), encoding=self.encoding)
        assert format_identifier == "ctxt"

        # next byte (u8) contains the "Common major version, currently 4"
        self.major_version = int.from_bytes(
            self.file.read(1), byteorder=self.byte_order, signed=self.signed
        )
        # next byte (u8) contains the "Specific minor version"
        self.minor_version = int.from_bytes(
            self.file.read(1), byteorder=self.byte_order, signed=self.signed
        )

        self.section_pointer = []
        self.section_size = []
        # In the header each section is given 16 bytes:
        #   - First 8 bytes specify the total size of the section (in bytes)
        #   - Last 8 bytes specify a pointer to the beggining of the section
        for i in range(len(self.read_order)):
            self.section_size.append(
                int.from_bytes(
                    self.file.read(8), byteorder=self.byte_order, signed=self.signed
                )
            )
            self.section_pointer.append(
                int.from_bytes(
                    self.file.read(8), byteorder=self.byte_order, signed=self.signed
                )
            )

    def __create_node_dict(
        self,
        identifier,
        profile_info,
        node_name,
        node,
        context_info,
        metric_names=False,
        value=0,
    ) -> None:
        """Stores profile and metric information for each context.
        node_dicts will be used later to create the dataframe."""
        self.node_dicts[identifier] = dict(profile_info)
        self.node_dicts[identifier]["name"] = node_name
        self.node_dicts[identifier]["node"] = node
        self.node_dicts[identifier].update(context_info)

        # create a metric value of 0 for all the missing values.
        if isinstance(metric_names, set):
            for metric in metric_names:
                self.node_dicts[identifier][metric] = value
        else:
            self.node_dicts[identifier][metric_names] = value

    def __read_cct_info_section(self, section_pointer: int, section_size: int) -> None:
        """
        Reader Context Trace Headers section of trace.db
        """
        # get to the right place in the file
        self.file.seek(section_pointer)

        # Header for each trace (u64)
        context_pointer = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )
        # Number of traces listed in this section (u32)
        self.file.read(4)
        # num_contexts = int.from_bytes(
        #     self.file.read(4), byteorder=self.byte_order, signed=self.signed
        # )
        # Size of a {CI} structure, currently 32
        context_size = int.from_bytes(
            self.file.read(1), byteorder=self.byte_order, signed=self.signed
        )

        # Read metric and profile information for each context.
        for i in self.meta_reader.context_map.keys():
            # get the corresponding Hatchet node for the HPCToolkit context.
            node = self.meta_reader.node_map[i]

            header_pointer = context_pointer + ((i) * context_size)
            self.__read_single_context_value_block(
                header_pointer,
                node,
                self.meta_reader.get_information_from_context_id(i),
            )

    def __read_single_context_value_block(
        self, header_pointer: int, node: Node, context_info: dict
    ) -> None:
        """
        Reads the value block that includes all the metrics and
        profile information (rank, thread, etc.) of a context.
        """

        def __read_profiles(
            next_metric_index,
            visited_profiles,
            node_name,
            node,
            context_info,
            metric_name,
        ):
            """
            Iterates over the profiles and corresponding metric values
            of a metric associated with a context.
            For example, in the first iteration, we focus on 'context1,'
            and within that context, we examine 'metric1.' Within
            'metric1,' we start with 'rank 0, thread 0,' and retrieve
            its associated value. In the second iteration of the while
            loop, we move to 'rank 0, thread 1' and retrieve its value.
            """
            while self.file.tell() < next_metric_index:
                prof_index = int.from_bytes(
                    self.file.read(4), byteorder=self.byte_order, signed=self.signed
                )
                value = struct.unpack("d", self.file.read(8))[0]

                # get the profile info by using its index.
                profile_info = self.profile_reader.get_hit_from_profile(prof_index)

                # keep track of the profiles that has no metric values.
                # we need to manually create these profiles to convert
                # HPCToolkit's sparse representation to the Hatchet
                # representation.
                visited_profiles.add(prof_index)

                # Example: (node, rank 0, thread 0)
                identifier = (node,) + tuple(profile_info.values())

                # add the metric value if the identifier has seen
                # before. Otherwise, add new information.
                if identifier in self.node_dicts.keys():
                    self.node_dicts[identifier][metric_name] = value
                else:
                    self.__create_node_dict(
                        identifier,
                        profile_info,
                        node_name,
                        node,
                        context_info,
                        metric_name,
                        value,
                    )

        self.file.seek(header_pointer)

        # Header for each trace (u64)
        nonzero_vals = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )

        # Stores the profile and corresponding metric value.
        profile_value_pairs = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )

        # Number of non-empty (i.e. non-zero) metrics.
        nonempty_metrics = int.from_bytes(
            self.file.read(2), byteorder=self.byte_order, signed=self.signed
        )

        # empty space
        self.file.read(6)

        metrics_to_index = int.from_bytes(
            self.file.read(8), byteorder=self.byte_order, signed=self.signed
        )

        metric_indices = []
        self.file.seek(metrics_to_index)
        for _ in range(nonempty_metrics):
            metric_id = int.from_bytes(
                self.file.read(2), byteorder=self.byte_order, signed=self.signed
            )
            start_index = int.from_bytes(
                self.file.read(8), byteorder=self.byte_order, signed=self.signed
            )
            metric_indices.append((metric_id, start_index))

        # construct the node name for the context. it will be
        # used to create the name column in the dataframe.
        node_name = None
        if "type" in context_info.keys():
            if context_info["type"] == "function":
                node_name = context_info["function"]
            # file:line
            elif context_info["type"] == "line":
                node_name = "{}:{}".format(context_info["file"], context_info["line"])
            # Loop@file:line
            elif context_info["type"] == "loop":
                node_name = "Loop@{}:{}".format(
                    context_info["file"], context_info["line"]
                )
            # module:instruction
            # Example: app.exe:0x1581
            elif context_info["type"] == "instruction":
                node_name = "{}:{}".format(
                    context_info["module"], context_info["instruction"]
                )
        # the root node, which is the parent of the main function,
        # don't have a type. its name is stored in the 'function' key.
        # that's why we create a 'type' for it.
        else:
            node_name = context_info["function"]
            context_info["type"] = "function"
            context_info["module"] = None
            context_info["file"] = None
            context_info.pop("loop_type", None)
            context_info.pop("lexical_type", None)

        # I encountered a context in which the 'function' was set to None:
        # {'module': '/usr/lib64/libmlx5.so.1.24.44.0',
        # 'file': '[libucs.so.0.0.0]',
        # 'function': None,
        # 'line': '-1',
        # 'type': 'function',
        # 'relation': 1,
        # 'instruction': '-0x1'}
        # I believe the type should have been 'instruction' instead of 'function'.
        # That might be a bug on the HPCToolkit side.
        if node_name is None:
            node_name = "{}:{}".format(
                context_info["module"], context_info["instruction"]
            )

        # add name to the node's frame.
        node.frame.attrs["name"] = node_name

        # we need to keep track of the visited profiles. HPCToolkit
        # sparse representation doesn't provide profile information
        # if it has only 0 values. We need to create them manually
        # to be able to show those profiles (i.e. rank/thread) on
        # the dataframe.
        visited_profiles = set([])
        not_visited_profiles = list(self.profile_reader.profile_info_list)
        visited_metrics = set([])

        # iterate over each metric.
        # context1 -> metric1, context1 -> metric2, ...
        for i in range(len(metric_indices)):
            metric_id = metric_indices[i][0]
            start_index = metric_indices[i][1]
            next_metric_index = None

            # find the correct location to read.
            # a profile and its value takes 12 bytes.
            if i != len(metric_indices) - 1:
                next_metric_index = profile_value_pairs + (
                    (metric_indices[i + 1][1]) * 12
                )
            else:
                next_metric_index = profile_value_pairs + ((nonzero_vals) * 12)
            self.file.seek(profile_value_pairs + (start_index * 12))

            # get the metric name and type (point, execution, function, lex-aware)
            # from the corresponding id.
            metric = self.meta_reader.metric_id_name_map[metric_id][0]
            metric_type = self.meta_reader.metric_id_name_map[metric_id][1]

            # store all the visited metrics.
            # we will use this later to fill the missing values.
            visited_metrics.add(metric)
            visited_metrics.add(metric + " (inc)")

            # get the default inclusive and exclusive metric type
            # for the corresponding metric.
            exc, inc = None, None
            if metric in self.defaults_reader.metric_defaults.keys():
                exc = self.defaults_reader.metric_defaults[metric]["exclusive"]
                inc = self.defaults_reader.metric_defaults[metric]["inclusive"]

            # Read profiles and values for the current metric.
            if metric_type == inc:
                metric = metric + " (inc)"
                self.inclusive_metrics.add(metric)
            elif metric_type == exc:
                self.exclusive_metrics.add(metric)

            # this function read profiles and keeps
            # track of visited profiles.
            __read_profiles(
                next_metric_index,
                visited_profiles,
                node_name,
                node,
                context_info,
                metric,
            )

        # remove the visited profiles.
        not_visited_profiles = [
            i for j, i in enumerate(not_visited_profiles) if j not in visited_profiles
        ]

        # iterate over the not visited nodes and create dummy instances.
        for profile in not_visited_profiles:
            hit_pointer = profile["hit_pointer"]
            if hit_pointer != 0:
                dummy_profile_info = self.profile_reader.hit_map[hit_pointer]
                dummy_identifier = (node,) + tuple(dummy_profile_info.values())

                # fills all the metric values in not visited profiles
                # with 0.
                self.__create_node_dict(
                    dummy_identifier,
                    dummy_profile_info,
                    node_name,
                    node,
                    context_info,
                    metric_names=visited_metrics,
                    value=0,
                )


class HPCToolkitV4Reader:
    def __init__(self, directory: str) -> None:
        self.directory = directory

    def read(self):
        self.meta_reader: MetaReader = MetaReader(self.directory + "/meta.db")
        self.profile_reader: ProfileReader = ProfileReader(
            self.directory + "/profile.db", self.meta_reader
        )
        self.defaults_reader: DefaultReader = DefaultReader(
            self.directory + "/metrics/default.yaml", self.meta_reader
        )
        self.cct_reader = CCTReader(
            self.directory + "/cct.db",
            self.meta_reader,
            self.profile_reader,
            self.defaults_reader,
        )

        return self.create_graphframe()

    def create_graphframe(self) -> GraphFrame:
        graph = Graph(self.meta_reader.roots)
        graph.enumerate_traverse()
        dataframe = pd.DataFrame.from_dict(data=self.cct_reader.node_dicts.values())

        columns = dataframe.columns
        if "thread" in columns and "rank" in columns:
            indices = ["node", "rank", "thread"]
        elif "rank" in columns:
            indices = ["node", "rank"]
        elif "thread" in columns:
            indices = ["node", "thread"]
        else:
            indices = ["node"]

        dataframe.set_index(indices, inplace=True)
        dataframe.sort_index(inplace=True)

        self.cct_reader.exclusive_metrics = list(self.cct_reader.exclusive_metrics)
        self.cct_reader.inclusive_metrics = list(self.cct_reader.inclusive_metrics)

        default_metric = "time"
        if "CPUTIME (sec)" in dataframe.columns:
            dataframe.rename(
                columns={
                    "CPUTIME (sec)": "time",
                    "CPUTIME (sec) (inc)": "time (inc)",
                },
                inplace=True,
            )
            self.cct_reader.exclusive_metrics.remove("CPUTIME (sec)")
            self.cct_reader.inclusive_metrics.remove("CPUTIME (sec) (inc)")
            self.cct_reader.exclusive_metrics.append("time")
            self.cct_reader.inclusive_metrics.append("time (inc)")
        else:
            default_metric = self.cct_reader.exclusive_metrics[0]

        # if the metric is numeric, fill NaNs with zero.
        dataframe[self.cct_reader.inclusive_metrics] = dataframe[
            self.cct_reader.inclusive_metrics
        ].fillna(0)
        dataframe[self.cct_reader.exclusive_metrics] = dataframe[
            self.cct_reader.exclusive_metrics
        ].fillna(0)

        if "line" in dataframe.columns:
            dataframe["line"] = dataframe["line"].astype("int64")
        if "core" in dataframe.columns:
            dataframe["core"] = dataframe["core"].astype("int64")
        if "node_pid" in dataframe.columns:
            dataframe["node_pid"] = dataframe["node_pid"].astype("int64")

        return hatchet.graphframe.GraphFrame(
            graph,
            dataframe,
            exc_metrics=self.cct_reader.exclusive_metrics,
            inc_metrics=self.cct_reader.inclusive_metrics,
            default_metric=default_metric,
        )



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd
import os

from hatchet import GraphFrame
from hatchet.readers.hpctoolkit_reader import HPCToolkitReader
from hatchet.external.console import ConsoleRenderer

modules = [
    "cpi",
    "/collab/usr/global/tools/hpctoolkit/chaos_5_x86_64_ib/"
    "hpctoolkit-2017-03-16/lib/hpctoolkit/ext-libs/libmonitor.so.0.0.0",
    "/usr/local/tools/mvapich2-intel-debug-2.2/lib/libmpi.so.12.0.5",
    "/lib64/libc-2.12.so",
    "/usr/lib64/libpsm_infinipath.so.1.14",
]

src_files = [
    "./src/cpi.c",
    "<unknown file>",
    "/tmp/dpkg-mkdeb.gouoc49UG7/src/mvapich/src/build/../src/mpi/" "init/init.c",
    "/tmp/dpkg-mkdeb.gouoc49UG7/src/mvapich/src/build/../src/mpi/" "init/initthread.c",
    "/tmp/dpkg-mkdeb.gouoc49UG7/src/mvapich/src/build/../src/mpid/"
    "ch3/src/mpid_init.c",
    "/tmp/dpkg-mkdeb.gouoc49UG7/src/mvapich/src/build/../src/mpid/"
    "ch3/channels/psm/src/mpidi_calls.c",
    "/tmp/dpkg-mkdeb.gouoc49UG7/src/mvapich/src/build/../src/mpid/"
    "ch3/channels/psm/src/psm_entry.c",
    "/tmp/dpkg-mkdeb.gouoc49UG7/src/mvapich/src/build/../src/mpi/" "init/finalize.c",
    "/tmp/dpkg-mkdeb.gouoc49UG7/src/mvapich/src/build/../src/mpid/"
    "ch3/src/mpid_finalize.c",
    "/tmp/dpkg-mkdeb.gouoc49UG7/src/mvapich/src/build/../src/mpid/"
    "ch3/channels/psm/src/psm_exit.c",
    "interp.c",
    "<unknown file>",
]

procedures = [
    "main",
    "<program root>",
    "MPI_Init",
    "pthread_create",
    "MPI_Finalize",
    "PMPI_Init",
    "MPIR_Init_thread",
    "MPID_Init",
    "MPIDI_CH3_Init",
    "MPIDI_CH3_Finalize",
    "psm_doinit",
    "PMPI_Finalize",
    "MPID_Finalize",
    "psm_dofinalize",
    "__GI_sched_yield",
    "<unknown procedure>",
]


def test_graphframe(data_dir, calc_pi_hpct_db):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))

    assert len(gf.dataframe.groupby("module")) == 5
    assert len(gf.dataframe.groupby("file")) == 11
    assert len(gf.dataframe.groupby("name")) == 20

    for col in gf.dataframe.columns:
        if col in ("time (inc)", "time"):
            assert gf.dataframe[col].dtype == np.float64
        elif col in ("nid", "rank", "line"):
            assert gf.dataframe[col].dtype == np.int64
        elif col in ("name", "type", "file", "module", "node"):
            assert gf.dataframe[col].dtype == object

    # add tests to confirm values in dataframe
    df = pd.read_csv(str(os.path.join(data_dir, "hpctoolkit-cpi-graphframe.csv")))

    gf.dataframe.reset_index(inplace=True)
    df.reset_index(inplace=True)

    gf.dataframe.sort_values(by=["nid", "rank"], inplace=True)
    df.sort_values(by=["nid", "rank"], inplace=True)

    t1 = gf.dataframe["time"].values
    t2 = df["time"].values

    ti1 = gf.dataframe["time (inc)"].values
    ti2 = df["time (inc)"].values

    for v1, v2 in zip(t1, t2):
        assert v1 == v2

    for v1, v2 in zip(ti1, ti2):
        assert v1 == v2


def test_tree(calc_pi_hpct_db):
    gf = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )
    assert "0.000 <program root> <unknown file>" in output
    assert (
        "0.000 198:MPIR_Init_thread /tmp/dpkg-mkdeb.gouoc49UG7/src/mvapich/src/build/../src/mpi/init/initthread.c"
        in output
    )

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time (inc)",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )
    assert "17989.000 interp.c:0 interp.c" in output
    assert (
        "999238.000 230:psm_dofinalize /tmp/dpkg-mkdeb.gouoc49UG7/src/mvapich/src/build/../src/mpid/ch3/channels/psm/src/psm_exit.c"
        in output
    )


def test_read_calc_pi_database(calc_pi_hpct_db):
    """Sanity check the HPCT database reader by examining a known input."""
    reader = HPCToolkitReader(str(calc_pi_hpct_db))
    reader.fill_tables()

    assert len(reader.load_modules) == 5
    assert len(reader.src_files) == 12
    assert len(reader.procedure_names) == 16
    assert all(lm in reader.load_modules.values() for lm in modules)
    assert all(sf in reader.src_files.values() for sf in src_files)
    assert all(pr in reader.procedure_names.values() for pr in procedures)


def test_allgather(data_dir, osu_allgather_hpct_db):
    gf = GraphFrame.from_hpctoolkit(str(osu_allgather_hpct_db))

    assert len(gf.dataframe.groupby("module")) == 9
    assert len(gf.dataframe.groupby("file")) == 41
    assert len(gf.dataframe.groupby("name")) == 383
    assert len(gf.dataframe.groupby("type")) == 3

    for col in gf.dataframe.columns:
        if col in ("time (inc)", "time"):
            assert gf.dataframe[col].dtype == np.float64
        elif col in ("nid", "rank", "thread", "line"):
            assert gf.dataframe[col].dtype == np.int64
        elif col in ("name", "type", "file", "module", "node"):
            assert gf.dataframe[col].dtype == object

    # add tests to confirm values in dataframe
    df = pd.read_csv(str(os.path.join(data_dir, "hpctoolkit-allgather-graphframe.csv")))

    gf.dataframe.reset_index(inplace=True)
    df.reset_index(inplace=True)

    gf.dataframe.sort_values(by=["nid", "rank", "thread"], inplace=True)
    df.sort_values(by=["nid", "rank", "thread"], inplace=True)

    t1 = gf.dataframe["time"].values
    t2 = df["time"].values

    ti1 = gf.dataframe["time (inc)"].values
    ti2 = df["time (inc)"].values

    for v1, v2 in zip(t1, t2):
        assert v1 == v2

    for v1, v2 in zip(ti1, ti2):
        assert v1 == v2


def test_graphframe_to_literal(data_dir, calc_pi_hpct_db):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))
    graph_literal = gf.to_literal()

    gf2 = GraphFrame.from_literal(graph_literal)

    assert len(gf.graph) == len(gf2.graph)


def test_graphframe_to_literal_with_threads(data_dir, osu_allgather_hpct_db):
    gf = GraphFrame.from_hpctoolkit(str(osu_allgather_hpct_db))
    graph_literal = gf.to_literal()

    gf2 = GraphFrame.from_literal(graph_literal)

    assert len(gf.graph) == len(gf2.graph)


def test_inclusive_time_calculation(data_dir, calc_pi_hpct_db):
    """Validate calculate_inclusive_metrics() on known dataset."""
    gf = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))

    # save original time (inc) column for correctness check
    gf.dataframe["orig_inc_time"] = gf.dataframe["time (inc)"]

    # remove original time (inc) column since it will be generated by calculate_inclusive_metrics()
    del gf.dataframe["time (inc)"]

    gf.calculate_inclusive_metrics()
    assert all(
        gf.dataframe["time (inc)"].values == gf.dataframe["orig_inc_time"].values
    )


# START OF TESTS FOR THE NEW FORMAT
def test_graphframe_v4(data_dir, calc_pi_hpct_v4_db):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_v4_db))

    for col in gf.dataframe.columns:
        if col in gf.inc_metrics or col in gf.exc_metrics:
            assert gf.dataframe[col].dtype == np.float64
        elif col in ("line"):
            assert gf.dataframe[col].dtype == np.int64
        elif col in ("name", "node", "file", "module"):
            assert gf.dataframe[col].dtype == object

    # TODO: add tests to confirm values in dataframe


def test_tree_v4(calc_pi_hpct_v4_db):
    gf = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_v4_db))

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time (inc)",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )
    assert "0.088 main src/home/ocankur/apps/test/hatchet_cpi/cpi.c" in output
    assert (
        "0.058 PMPI_Reduce [libmpi.so.40.30.1] src/home/ocankur/apps/test/hatchet_cpi/cpi.c"
        in output
    )
    assert "0.029 MPI_Finalize src/home/ocankur/apps/test/hatchet_cpi/cpi.c" in output
    assert (
        "0.000 PMPI_Bcast [libmpi.so.40.30.1] src/home/ocankur/apps/test/hatchet_cpi/cpi.c"
        in output
    )

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time (inc)",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=3,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )
    assert "0.090 main src/home/ocankur/apps/test/hatchet_cpi/cpi.c" in output
    assert (
        "0.059 PMPI_Bcast [libmpi.so.40.30.1] src/home/ocankur/apps/test/hatchet_cpi/cpi.c"
        in output
    )
    assert (
        "0.000 PMPI_Reduce [libmpi.so.40.30.1] src/home/ocankur/apps/test/hatchet_cpi/cpi.c"
        in output
    )
    assert "0.030 MPI_Finalize src/home/ocankur/apps/test/hatchet_cpi/cpi.c" in output


# def test_inclusive_metric_calculation_new_format(data_dir, saxpy_mpi_cuda_newhpct_db):
#     """Validate calculate_inclusive_metrics() on known dataset."""
#     gf = GraphFrame.from_hpctoolkit(str(saxpy_mpi_cuda_newhpct_db))

#     metric = gf.exc_metrics[0]
#     # save original time (inc) column for correctness check
#     gf.dataframe["orig_inc_time"] = gf.dataframe[metric + " (inc)"]

#     # remove original time (inc) column since it will be generated by calculate_inclusive_metrics()
#     del gf.dataframe[metric + " (inc)"]

#     gf.calculate_inclusive_metrics()
#     assert all(
#         gf.dataframe[metric + " (inc)"].values == gf.dataframe["orig_inc_time"].values
#     )


def test_graphframe_to_literal__v4(data_dir, calc_pi_hpct_v4_db):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_v4_db))
    graph_literal = gf.to_literal()

    gf2 = GraphFrame.from_literal(graph_literal)

    assert len(gf.graph) == len(gf2.graph)
